def fill_price_worse_than_limit_price(fill_price, order):
    """Checks whether the fill price is worse than the order's limit price.

    Parameters
    ----------
    fill_price: float
        The price to check.

    order: ziplime.finance.order.Order
        The order whose limit price to check.

    Returns
    -------
    bool: Whether the fill price is above the limit price (for a buy) or below
    the limit price (for a sell).
    """
    if order.limit:
        # this is tricky! if an order with a limit price has reached
        # the limit price, we will try to fill the order. do not fill
        # these shares if the impacted price is worse than the limit
        # price. return early to avoid creating the transaction.

        # buy order is worse if the impacted price is greater than
        # the limit price. sell order is worse if the impacted price
        # is less than the limit price
        if (order.direction > 0 and fill_price > order.limit) or (
                order.direction < 0 and fill_price < order.limit
        ):
            return True

    return False



"""migration

Revision ID: 498bbf0b3be2
Revises: 
Create Date: 2025-09-15 10:09:29.906049

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '498bbf0b3be2'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('asset_router',
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('asset_type', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('sid')
    )
    op.create_table('dividend_payouts',
    sa.Column('index', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('ex_date', sa.Date(), nullable=False),
    sa.Column('declared_date', sa.Date(), nullable=False),
    sa.Column('record_date', sa.Date(), nullable=False),
    sa.Column('pay_date', sa.Date(), nullable=False),
    sa.Column('amount', sa.Integer(), nullable=False),
    sa.PrimaryKeyConstraint('index')
    )
    op.create_index(op.f('ix_dividend_payouts_ex_date'), 'dividend_payouts', ['ex_date'], unique=False)
    op.create_index(op.f('ix_dividend_payouts_sid'), 'dividend_payouts', ['sid'], unique=False)
    op.create_table('dividends',
    sa.Column('index', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('effective_date', sa.Date(), nullable=False),
    sa.Column('ratio', sa.Float(), nullable=False),
    sa.PrimaryKeyConstraint('index')
    )
    op.create_index(op.f('ix_dividends_effective_date'), 'dividends', ['effective_date'], unique=False)
    op.create_index(op.f('ix_dividends_sid'), 'dividends', ['sid'], unique=False)
    op.create_table('exchanges',
    sa.Column('exchange', sa.String(), nullable=False),
    sa.Column('canonical_name', sa.String(), nullable=False),
    sa.Column('country_code', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('exchange')
    )
    op.create_table('mergers',
    sa.Column('index', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('effective_date', sa.Date(), nullable=False),
    sa.Column('ratio', sa.Float(), nullable=False),
    sa.PrimaryKeyConstraint('index')
    )
    op.create_index(op.f('ix_mergers_effective_date'), 'mergers', ['effective_date'], unique=False)
    op.create_index(op.f('ix_mergers_sid'), 'mergers', ['sid'], unique=False)
    op.create_table('splits',
    sa.Column('index', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('effective_date', sa.Date(), nullable=False),
    sa.Column('ratio', sa.Float(), nullable=False),
    sa.PrimaryKeyConstraint('index')
    )
    op.create_index(op.f('ix_splits_effective_date'), 'splits', ['effective_date'], unique=False)
    op.create_index(op.f('ix_splits_sid'), 'splits', ['sid'], unique=False)
    op.create_table('stock_dividend_payouts',
    sa.Column('index', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('ex_date', sa.Date(), nullable=False),
    sa.Column('declared_date', sa.Date(), nullable=False),
    sa.Column('record_date', sa.Date(), nullable=False),
    sa.Column('pay_date', sa.Date(), nullable=False),
    sa.Column('payment_sid', sa.String(), nullable=False),
    sa.Column('ration', sa.Float(), nullable=False),
    sa.PrimaryKeyConstraint('index')
    )
    op.create_index(op.f('ix_stock_dividend_payouts_ex_date'), 'stock_dividend_payouts', ['ex_date'], unique=False)
    op.create_index(op.f('ix_stock_dividend_payouts_sid'), 'stock_dividend_payouts', ['sid'], unique=False)
    op.create_table('symbols_universe',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(), nullable=False),
    sa.Column('symbol', sa.String(), nullable=False),
    sa.Column('universe_type', sa.String(), nullable=False),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_table('commodities',
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('mic', sa.String(), nullable=True),
    sa.Column('asset_name', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('first_traded', sa.Date(), nullable=False),
    sa.Column('auto_close_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['sid'], ['asset_router.sid'], ),
    sa.PrimaryKeyConstraint('sid')
    )
    op.create_table('currencies',
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('mic', sa.String(), nullable=True),
    sa.Column('asset_name', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('first_traded', sa.Date(), nullable=False),
    sa.Column('auto_close_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['sid'], ['asset_router.sid'], ),
    sa.PrimaryKeyConstraint('sid')
    )
    op.create_table('equities',
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('mic', sa.String(), nullable=True),
    sa.Column('asset_name', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('first_traded', sa.Date(), nullable=False),
    sa.Column('auto_close_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['sid'], ['asset_router.sid'], ),
    sa.PrimaryKeyConstraint('sid')
    )
    op.create_table('futures_root_symbols',
    sa.Column('root_symbol', sa.String(), nullable=False),
    sa.Column('root_symbol_id', sa.Integer(), nullable=False),
    sa.Column('sector', sa.String(), nullable=False),
    sa.Column('description', sa.String(), nullable=False),
    sa.Column('exchange', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['exchange'], ['exchanges.exchange'], ),
    sa.PrimaryKeyConstraint('root_symbol')
    )
    op.create_index(op.f('ix_futures_root_symbols_exchange'), 'futures_root_symbols', ['exchange'], unique=False)
    op.create_table('symbols_universe_assets',
    sa.Column('symbol_universe_id', sa.Integer(), nullable=False),
    sa.Column('asset_sid', sa.Integer(), nullable=False),
    sa.Column('ratio', sa.Numeric(precision=12, scale=8), nullable=True),
    sa.ForeignKeyConstraint(['asset_sid'], ['asset_router.sid'], ),
    sa.ForeignKeyConstraint(['symbol_universe_id'], ['symbols_universe.id'], ),
    sa.PrimaryKeyConstraint('symbol_universe_id', 'asset_sid')
    )
    op.create_table('currency_symbol_mappings',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('symbol', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('exchange', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['exchange'], ['exchanges.exchange'], ),
    sa.ForeignKeyConstraint(['sid'], ['currencies.sid'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_currency_symbol_mappings_exchange'), 'currency_symbol_mappings', ['exchange'], unique=False)
    op.create_index(op.f('ix_currency_symbol_mappings_sid'), 'currency_symbol_mappings', ['sid'], unique=False)
    op.create_table('equity_supplementary_mappings',
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('field', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('value', sa.String(), nullable=False),
    sa.ForeignKeyConstraint(['sid'], ['equities.sid'], ),
    sa.PrimaryKeyConstraint('sid', 'field', 'start_date')
    )
    op.create_index(op.f('ix_equity_supplementary_mappings_end_date'), 'equity_supplementary_mappings', ['end_date'], unique=False)
    op.create_index(op.f('ix_equity_supplementary_mappings_field'), 'equity_supplementary_mappings', ['field'], unique=False)
    op.create_index(op.f('ix_equity_supplementary_mappings_sid'), 'equity_supplementary_mappings', ['sid'], unique=False)
    op.create_index(op.f('ix_equity_supplementary_mappings_start_date'), 'equity_supplementary_mappings', ['start_date'], unique=False)
    op.create_table('equity_symbol_mappings',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('symbol', sa.String(), nullable=False),
    sa.Column('company_symbol', sa.String(), nullable=False),
    sa.Column('share_class_symbol', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('exchange', sa.String(), nullable=True),
    sa.ForeignKeyConstraint(['exchange'], ['exchanges.exchange'], ),
    sa.ForeignKeyConstraint(['sid'], ['equities.sid'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.create_index(op.f('ix_equity_symbol_mappings_company_symbol'), 'equity_symbol_mappings', ['company_symbol'], unique=False)
    op.create_index(op.f('ix_equity_symbol_mappings_exchange'), 'equity_symbol_mappings', ['exchange'], unique=False)
    op.create_index(op.f('ix_equity_symbol_mappings_sid'), 'equity_symbol_mappings', ['sid'], unique=False)
    op.create_table('futures_contracts',
    sa.Column('symbol', sa.String(), nullable=False),
    sa.Column('root_symbol', sa.String(), nullable=False),
    sa.Column('notice_date', sa.Date(), nullable=False),
    sa.Column('expiration_date', sa.Date(), nullable=False),
    sa.Column('multiplier', sa.Float(), nullable=False),
    sa.Column('tick_size', sa.Float(), nullable=False),
    sa.Column('exchange', sa.String(), nullable=False),
    sa.Column('sid', sa.Integer(), nullable=False),
    sa.Column('mic', sa.String(), nullable=True),
    sa.Column('asset_name', sa.String(), nullable=False),
    sa.Column('start_date', sa.Date(), nullable=False),
    sa.Column('end_date', sa.Date(), nullable=False),
    sa.Column('first_traded', sa.Date(), nullable=False),
    sa.Column('auto_close_date', sa.Date(), nullable=False),
    sa.ForeignKeyConstraint(['exchange'], ['exchanges.exchange'], ),
    sa.ForeignKeyConstraint(['root_symbol'], ['futures_root_symbols.root_symbol'], ),
    sa.ForeignKeyConstraint(['sid'], ['asset_router.sid'], ),
    sa.PrimaryKeyConstraint('sid'),
    sa.UniqueConstraint('symbol')
    )
    op.create_index(op.f('ix_futures_contracts_exchange'), 'futures_contracts', ['exchange'], unique=False)
    op.create_index(op.f('ix_futures_contracts_root_symbol'), 'futures_contracts', ['root_symbol'], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f('ix_futures_contracts_root_symbol'), table_name='futures_contracts')
    op.drop_index(op.f('ix_futures_contracts_exchange'), table_name='futures_contracts')
    op.drop_table('futures_contracts')
    op.drop_index(op.f('ix_equity_symbol_mappings_sid'), table_name='equity_symbol_mappings')
    op.drop_index(op.f('ix_equity_symbol_mappings_exchange'), table_name='equity_symbol_mappings')
    op.drop_index(op.f('ix_equity_symbol_mappings_company_symbol'), table_name='equity_symbol_mappings')
    op.drop_table('equity_symbol_mappings')
    op.drop_index(op.f('ix_equity_supplementary_mappings_start_date'), table_name='equity_supplementary_mappings')
    op.drop_index(op.f('ix_equity_supplementary_mappings_sid'), table_name='equity_supplementary_mappings')
    op.drop_index(op.f('ix_equity_supplementary_mappings_field'), table_name='equity_supplementary_mappings')
    op.drop_index(op.f('ix_equity_supplementary_mappings_end_date'), table_name='equity_supplementary_mappings')
    op.drop_table('equity_supplementary_mappings')
    op.drop_index(op.f('ix_currency_symbol_mappings_sid'), table_name='currency_symbol_mappings')
    op.drop_index(op.f('ix_currency_symbol_mappings_exchange'), table_name='currency_symbol_mappings')
    op.drop_table('currency_symbol_mappings')
    op.drop_table('symbols_universe_assets')
    op.drop_index(op.f('ix_futures_root_symbols_exchange'), table_name='futures_root_symbols')
    op.drop_table('futures_root_symbols')
    op.drop_table('equities')
    op.drop_table('currencies')
    op.drop_table('commodities')
    op.drop_table('symbols_universe')
    op.drop_index(op.f('ix_stock_dividend_payouts_sid'), table_name='stock_dividend_payouts')
    op.drop_index(op.f('ix_stock_dividend_payouts_ex_date'), table_name='stock_dividend_payouts')
    op.drop_table('stock_dividend_payouts')
    op.drop_index(op.f('ix_splits_sid'), table_name='splits')
    op.drop_index(op.f('ix_splits_effective_date'), table_name='splits')
    op.drop_table('splits')
    op.drop_index(op.f('ix_mergers_sid'), table_name='mergers')
    op.drop_index(op.f('ix_mergers_effective_date'), table_name='mergers')
    op.drop_table('mergers')
    op.drop_table('exchanges')
    op.drop_index(op.f('ix_dividends_sid'), table_name='dividends')
    op.drop_index(op.f('ix_dividends_effective_date'), table_name='dividends')
    op.drop_table('dividends')
    op.drop_index(op.f('ix_dividend_payouts_sid'), table_name='dividend_payouts')
    op.drop_index(op.f('ix_dividend_payouts_ex_date'), table_name='dividend_payouts')
    op.drop_table('dividend_payouts')
    op.drop_table('asset_router')
    # ### end Alembic commands ###



import datetime

import polars as pl

from ziplime.constants.period import Period


class DataBundleSource:

    def __init__(self):
        pass

    async def get_data(self, symbols: list[str],
                       frequency: datetime.timedelta | Period,
                       date_from: datetime.datetime,
                       date_to: datetime.datetime,
                       **kwargs
                       ) -> pl.DataFrame:
        pass

    def get_data_sync(self, symbols: list[str],
                      frequency: datetime.timedelta | Period,
                      date_from: datetime.datetime,
                      date_to: datetime.datetime,
                      ) -> pl.DataFrame:
        pass



import datetime
from typing import Any

import pandas as pd
from exchange_calendars import ExchangeCalendar


from ziplime.exchanges.exchange import Exchange
from ziplime.finance.domain.ledger import Ledger
from ziplime.sources.benchmark_source import BenchmarkSource



class MaxLeverage:
    """Tracks the maximum account leverage."""

    def start_of_simulation(self, ledger: Ledger, emission_rate: datetime.timedelta, trading_calendar: ExchangeCalendar,
                            sessions: pd.DatetimeIndex, benchmark_source: BenchmarkSource):
        self._max_leverage = 0.0

    def end_of_bar(self, packet: dict[str, Any], ledger: Ledger, session: datetime.datetime, session_ix: int,
                   exchanges: dict[str, Exchange]):
        self._max_leverage = max(self._max_leverage, ledger.account.leverage)
        packet["cumulative_risk_metrics"]["max_leverage"] = self._max_leverage

    end_of_session = end_of_bar



from ziplime.finance.controls.account_control import AccountControl


class MaxLeverage(AccountControl):
    """AccountControl representing a limit on the maximum leverage allowed
    by the algorithm.
    """

    def __init__(self, max_leverage):
        """max_leverage is the gross leverage in decimal form. For example,
        2, limits an algorithm to trading at most double the account value.
        """
        super(MaxLeverage, self).__init__(max_leverage=max_leverage)
        self.max_leverage = max_leverage

        if max_leverage is None:
            raise ValueError("Must supply max_leverage")

        if max_leverage < 0:
            raise ValueError("max_leverage must be positive")

    def validate(self, _portfolio, _account, _algo_datetime, _algo_current_data):
        """Fail if the leverage is greater than the allowed leverage."""
        if _account.leverage > self.max_leverage:
            self.fail()



"""Simple common factors.
"""

from numbers import Number
from numpy import (
    arange,
    average,
    clip,
    copyto,
    exp,
    fmax,
    full,
    isnan,
    log,
    # NINF,
    sqrt,
    sum as np_sum,
    unique,
    errstate as np_errstate,
)
import numpy as np

from ziplime.pipeline.data import EquityPricing
from ziplime.utils.math_utils import (
    nanargmax,
    nanmax,
    nanmean,
    nanstd,
    nansum,
)
from ziplime.utils.numpy_utils import (
    float64_dtype,
    # ignore_nanwarnings,
)

from .factor import CustomFactor
from ..mixins import SingleInputMixin


class Returns(CustomFactor):
    """
    Calculates the percent change in close price over the given window_length.

    **Default Inputs**: [EquityPricing.close]
    """

    inputs = [EquityPricing.close]
    window_safe = True

    def _validate(self):
        super(Returns, self)._validate()
        if self.window_length < 2:
            raise ValueError(
                "'Returns' expected a window length of at least 2, but was "
                "given {window_length}. For daily returns, use a window "
                "length of 2.".format(window_length=self.window_length)
            )

    def compute(self, today, assets, out, close):
        out[:] = (close[-1] - close[0]) / close[0]


class PercentChange(SingleInputMixin, CustomFactor):
    """
    Calculates the percent change over the given window_length.

    **Default Inputs:** None

    **Default Window Length:** None

    Notes
    -----
    Percent change is calculated as ``(new - old) / abs(old)``.
    """

    window_safe = True

    def _validate(self):
        super(PercentChange, self)._validate()
        if self.window_length < 2:
            raise ValueError(
                "'PercentChange' expected a window length"
                "of at least 2, but was given {window_length}. "
                "For daily percent change, use a window "
                "length of 2.".format(window_length=self.window_length)
            )

    def compute(self, today, assets, out, values):
        with np_errstate(divide="ignore", invalid="ignore"):
            out[:] = (values[-1] - values[0]) / abs(values[0])


class DailyReturns(Returns):
    """
    Calculates daily percent change in close price.

    **Default Inputs**: [EquityPricing.close]
    """

    inputs = [EquityPricing.close]
    window_safe = True
    window_length = 2


class SimpleMovingAverage(SingleInputMixin, CustomFactor):
    """
    Average Value of an arbitrary column

    **Default Inputs**: None

    **Default Window Length**: None
    """

    # numpy's nan functions throw warnings when passed an array containing only
    # nans, but they still returns the desired value (nan), so we ignore the
    # warning.
    # ctx = ignore_nanwarnings()

    def compute(self, today, assets, out, data):
        out[:] = nanmean(data, axis=0)


class WeightedAverageValue(CustomFactor):
    """
    Helper for VWAP-like computations.

    **Default Inputs:** None

    **Default Window Length:** None
    """

    def compute(self, today, assets, out, base, weight):
        out[:] = nansum(base * weight, axis=0) / nansum(weight, axis=0)


class VWAP(WeightedAverageValue):
    """
    Volume Weighted Average Price

    **Default Inputs:** [EquityPricing.close, EquityPricing.volume]

    **Default Window Length:** None
    """

    inputs = (EquityPricing.close, EquityPricing.volume)


class MaxDrawdown(SingleInputMixin, CustomFactor):
    """
    Max Drawdown

    **Default Inputs:** None

    **Default Window Length:** None
    """

    # ctx = ignore_nanwarnings()

    def compute(self, today, assets, out, data):
        drawdowns = fmax.accumulate(data, axis=0) - data
        drawdowns[isnan(drawdowns)] = -np.inf
        drawdown_ends = nanargmax(drawdowns, axis=0)

        # TODO: Accelerate this loop in Cython or Numba.
        for i, end in enumerate(drawdown_ends):
            peak = nanmax(data[: end + 1, i])
            out[i] = (peak - data[end, i]) / data[end, i]


class AverageDollarVolume(CustomFactor):
    """
    Average Daily Dollar Volume

    **Default Inputs:** [EquityPricing.close, EquityPricing.volume]

    **Default Window Length:** None
    """

    inputs = [EquityPricing.close, EquityPricing.volume]

    def compute(self, today, assets, out, close, volume):
        out[:] = nansum(close * volume, axis=0) / len(close)


def exponential_weights(length, decay_rate):
    """
    Build a weight vector for an exponentially-weighted statistic.

    The resulting ndarray is of the form::

        [decay_rate ** length, ..., decay_rate ** 2, decay_rate]

    Parameters
    ----------
    length : int
        The length of the desired weight vector.
    decay_rate : float
        The rate at which entries in the weight vector increase or decrease.

    Returns
    -------
    weights : ndarray[float64]
    """
    return full(length, decay_rate, float64_dtype) ** arange(length + 1, 1, -1)


class _ExponentialWeightedFactor(SingleInputMixin, CustomFactor):
    """
    Base class for factors implementing exponential-weighted operations.

    **Default Inputs:** None

    **Default Window Length:** None

    Parameters
    ----------
    inputs : length-1 list or tuple of BoundColumn
        The expression over which to compute the average.
    window_length : int > 0
        Length of the lookback window over which to compute the average.
    decay_rate : float, 0 < decay_rate <= 1
        Weighting factor by which to discount past observations.

        When calculating historical averages, rows are multiplied by the
        sequence::

            decay_rate, decay_rate ** 2, decay_rate ** 3, ...

    Methods
    -------
    weights
    from_span
    from_halflife
    from_center_of_mass
    """

    params = ("decay_rate",)

    @classmethod
    def from_span(cls, inputs, window_length, span: Number, **kwargs):
        """
        Convenience constructor for passing `decay_rate` in terms of `span`.

        Forwards `decay_rate` as `1 - (2.0 / (1 + span))`.  This provides the
        behavior equivalent to passing `span` to pandas.ewma.

        Examples
        --------
        .. code-block:: python

            # Equivalent to:
            # my_ewma = EWMA(
            #    inputs=[EquityPricing.close],
            #    window_length=30,
            #    decay_rate=(1 - (2.0 / (1 + 15.0))),
            # )
            my_ewma = EWMA.from_span(
                inputs=[EquityPricing.close],
                window_length=30,
                span=15,
            )

        Notes
        -----
        This classmethod is provided by both
        :class:`ExponentialWeightedMovingAverage` and
        :class:`ExponentialWeightedMovingStdDev`.
        """
        if span <= 1:
            raise ValueError("`span` must be a positive number. %s was passed." % span)

        decay_rate = 1.0 - (2.0 / (1.0 + span))
        assert 0.0 < decay_rate <= 1.0

        return cls(
            inputs=inputs, window_length=window_length, decay_rate=decay_rate, **kwargs
        )

    @classmethod
    def from_halflife(cls, inputs, window_length, halflife: Number, **kwargs):
        """
        Convenience constructor for passing ``decay_rate`` in terms of half
        life.

        Forwards ``decay_rate`` as ``exp(log(.5) / halflife)``.  This provides
        the behavior equivalent to passing `halflife` to pandas.ewma.

        Examples
        --------
        .. code-block:: python

            # Equivalent to:
            # my_ewma = EWMA(
            #    inputs=[EquityPricing.close],
            #    window_length=30,
            #    decay_rate=np.exp(np.log(0.5) / 15),
            # )
            my_ewma = EWMA.from_halflife(
                inputs=[EquityPricing.close],
                window_length=30,
                halflife=15,
            )

        Notes
        -----
        This classmethod is provided by both
        :class:`ExponentialWeightedMovingAverage` and
        :class:`ExponentialWeightedMovingStdDev`.
        """
        if halflife <= 0:
            raise ValueError(
                "`span` must be a positive number. %s was passed." % halflife
            )
        decay_rate = exp(log(0.5) / halflife)
        assert 0.0 < decay_rate <= 1.0

        return cls(
            inputs=inputs, window_length=window_length, decay_rate=decay_rate, **kwargs
        )

    @classmethod
    def from_center_of_mass(cls, inputs, window_length, center_of_mass, **kwargs):
        """
        Convenience constructor for passing `decay_rate` in terms of center of
        mass.

        Forwards `decay_rate` as `1 - (1 / 1 + center_of_mass)`.  This provides
        behavior equivalent to passing `center_of_mass` to pandas.ewma.

        Examples
        --------
        .. code-block:: python

            # Equivalent to:
            # my_ewma = EWMA(
            #    inputs=[EquityPricing.close],
            #    window_length=30,
            #    decay_rate=(1 - (1 / 15.0)),
            # )
            my_ewma = EWMA.from_center_of_mass(
                inputs=[EquityPricing.close],
                window_length=30,
                center_of_mass=15,
            )

        Notes
        -----
        This classmethod is provided by both
        :class:`ExponentialWeightedMovingAverage` and
        :class:`ExponentialWeightedMovingStdDev`.
        """
        return cls(
            inputs=inputs,
            window_length=window_length,
            decay_rate=(1.0 - (1.0 / (1.0 + center_of_mass))),
            **kwargs,
        )


class ExponentialWeightedMovingAverage(_ExponentialWeightedFactor):
    """
    Exponentially Weighted Moving Average

    **Default Inputs:** None

    **Default Window Length:** None

    Parameters
    ----------
    inputs : length-1 list/tuple of BoundColumn
        The expression over which to compute the average.
    window_length : int > 0
        Length of the lookback window over which to compute the average.
    decay_rate : float, 0 < decay_rate <= 1
        Weighting factor by which to discount past observations.

        When calculating historical averages, rows are multiplied by the
        sequence::

            decay_rate, decay_rate ** 2, decay_rate ** 3, ...

    Notes
    -----
    - This class can also be imported under the name ``EWMA``.

    See Also
    --------
    :meth:`pandas.DataFrame.ewm`
    """

    def compute(self, today, assets, out, data, decay_rate):
        out[:] = average(
            data,
            axis=0,
            weights=exponential_weights(len(data), decay_rate),
        )


class ExponentialWeightedMovingStdDev(_ExponentialWeightedFactor):
    """
    Exponentially Weighted Moving Standard Deviation

    **Default Inputs:** None

    **Default Window Length:** None

    Parameters
    ----------
    inputs : length-1 list/tuple of BoundColumn
        The expression over which to compute the average.
    window_length : int > 0
        Length of the lookback window over which to compute the average.
    decay_rate : float, 0 < decay_rate <= 1
        Weighting factor by which to discount past observations.

        When calculating historical averages, rows are multiplied by the
        sequence::

            decay_rate, decay_rate ** 2, decay_rate ** 3, ...

    Notes
    -----
    - This class can also be imported under the name ``EWMSTD``.

    See Also
    --------
    :func:`pandas.DataFrame.ewm`
    """

    def compute(self, today, assets, out, data, decay_rate):
        weights = exponential_weights(len(data), decay_rate)

        mean = average(data, axis=0, weights=weights)
        variance = average((data - mean) ** 2, axis=0, weights=weights)

        squared_weight_sum = np_sum(weights) ** 2
        bias_correction = squared_weight_sum / (squared_weight_sum - np_sum(weights**2))
        out[:] = sqrt(variance * bias_correction)


class LinearWeightedMovingAverage(SingleInputMixin, CustomFactor):
    """
    Weighted Average Value of an arbitrary column

    **Default Inputs**: None

    **Default Window Length**: None
    """

    # numpy's nan functions throw warnings when passed an array containing only
    # nans, but they still returns the desired value (nan), so we ignore the
    # warning.
    # ctx = ignore_nanwarnings()

    def compute(self, today, assets, out, data):
        ndays = data.shape[0]

        # Initialize weights array
        weights = arange(1, ndays + 1, dtype=float64_dtype).reshape(ndays, 1)

        # Compute normalizer
        normalizer = (ndays * (ndays + 1)) / 2

        # Weight the data
        weighted_data = data * weights

        # Compute weighted averages
        out[:] = nansum(weighted_data, axis=0) / normalizer


class AnnualizedVolatility(CustomFactor):
    """
    Volatility. The degree of variation of a series over time as measured by
    the standard deviation of daily returns.
    https://en.wikipedia.org/wiki/Volatility_(finance)

    **Default Inputs:** [Returns(window_length=2)]

    Parameters
    ----------
    annualization_factor : float, optional
        The number of time units per year. Defaults is 252, the number of NYSE
        trading days in a normal year.
    """

    inputs = [Returns(window_length=2)]
    params = {"annualization_factor": 252.0}
    window_length = 252

    def compute(self, today, assets, out, returns, annualization_factor):
        out[:] = nanstd(returns, axis=0) * (annualization_factor**0.5)


class PeerCount(SingleInputMixin, CustomFactor):
    """
    Peer Count of distinct categories in a given classifier.  This factor
    is returned by the classifier instance method peer_count()

    **Default Inputs:** None

    **Default Window Length:** 1
    """

    window_length = 1

    def _validate(self):
        super(PeerCount, self)._validate()
        if self.window_length != 1:
            raise ValueError(
                "'PeerCount' expected a window length of 1, but was given"
                "{window_length}.".format(window_length=self.window_length)
            )

    def compute(self, today, assets, out, classifier_values):
        # Convert classifier array to group label int array
        group_labels, null_label = self.inputs[0]._to_integral(classifier_values[0])
        _, inverse, counts = unique(  # Get counts, idx of unique groups
            group_labels,
            return_counts=True,
            return_inverse=True,
        )
        copyto(out, counts[inverse], where=(group_labels != null_label))


# Convenience aliases
EWMA = ExponentialWeightedMovingAverage
EWMSTD = ExponentialWeightedMovingStdDev


class Clip(CustomFactor):
    """
    Clip (limit) the values in a factor.

    Given an interval, values outside the interval are clipped to the interval
    edges. For example, if an interval of ``[0, 1]`` is specified, values
    smaller than 0 become 0, and values larger than 1 become 1.

    **Default Window Length:** 1

    Parameters
    ----------
    min_bound : float
        The minimum value to use.
    max_bound : float
        The maximum value to use.

    Notes
    -----
    To only clip values on one side, ``-np.inf` and ``np.inf`` may be passed.
    For example, to only clip the maximum value but not clip a minimum value:

    .. code-block:: python

       Clip(inputs=[factor], min_bound=-np.inf, max_bound=user_provided_max)

    See Also
    --------
    numpy.clip
    """

    window_length = 1
    params = ("min_bound", "max_bound")

    def compute(self, today, assets, out, values, min_bound, max_bound):
        clip(values[-1], min_bound, max_bound, out=out)



import datetime

from ziplime.trading.entities.orders.order_request import OrderRequest
from ziplime.trading.entities.trading_pair import TradingPair
from ziplime.trading.enums.order_side import OrderSide
from ziplime.trading.enums.order_type import OrderType


class MarketOrderRequest(OrderRequest):

    def __init__(self, order_id: str, trading_pair: TradingPair, order_side: OrderSide, quantity: float,
                 exchange_name: str, creation_date: datetime.datetime
                 ):
        super().__init__(order_id=order_id, trading_pair=trading_pair, exchange_name=exchange_name,
                         creation_date=exchange_name, quantity=quantity,
                         order_type=OrderType.MARKET)
        self.order_side = order_side
        self.quantity = quantity
        self.trading_pair = trading_pair



import abc
from collections import namedtuple, OrderedDict
from itertools import repeat
from textwrap import dedent
from weakref import WeakKeyDictionary

from toolz import first
from ziplime.currency import Currency
from ziplime.data.fx import DEFAULT_FX_RATE

from ziplime.pipeline.classifiers import Classifier, Latest as LatestClassifier
from ziplime.pipeline.domain import Domain, GENERIC
from ziplime.pipeline.factors import Factor, Latest as LatestFactor
from ziplime.pipeline.filters import Filter, Latest as LatestFilter
from ziplime.pipeline.term import (
    AssetExists,
    LoadableTerm,
    validate_dtype,
)
from ziplime.utils.formatting import s, plural
from ziplime.utils.numpy_utils import float64_dtype, NoDefaultMissingValue
from ziplime.utils.string_formatting import bulleted_list

IsSpecialization = "IsSpecialization"


class Column:
    """
    An abstract column of data, not yet associated with a dataset.
    """

    def __init__(
            self,
            dtype,
            missing_value=None,
            doc=None,
            metadata=None,
            currency_aware=False,
    ):
        if currency_aware and dtype != float64_dtype:
            raise ValueError(
                "Columns cannot be constructed with currency_aware={}, "
                "dtype={}. Currency aware columns must have a float64 dtype.".format(
                    currency_aware, dtype
                )
            )

        self.dtype = dtype
        self.missing_value = missing_value
        self.doc = doc
        self.metadata = metadata.copy() if metadata is not None else {}
        self.currency_aware = currency_aware

    def bind(self, name):
        """
        Bind a `Column` object to its name.
        """
        return _BoundColumnDescr(
            dtype=self.dtype,
            missing_value=self.missing_value,
            name=name,
            doc=self.doc,
            metadata=self.metadata,
            currency_aware=self.currency_aware,
        )


class _BoundColumnDescr:
    """
    Intermediate class that sits on `DataSet` objects and returns memoized
    `BoundColumn` objects when requested.

    This exists so that subclasses of DataSets don't share columns with their
    parent classes.
    """

    def __init__(self, dtype, missing_value, name, doc, metadata, currency_aware):
        # Validating and calculating default missing values here guarantees
        # that we fail quickly if the user passes an unsupporte dtype or fails
        # to provide a missing value for a dtype that requires one
        # (e.g. int64), but still enables us to provide an error message that
        # points to the name of the failing column.
        try:
            self.dtype, self.missing_value = validate_dtype(
                termname="Column(name={name!r})".format(name=name),
                dtype=dtype,
                missing_value=missing_value,
            )
        except NoDefaultMissingValue as exc:
            # Re-raise with a more specific message.
            raise NoDefaultMissingValue(
                "Failed to create Column with name {name!r} and"
                " dtype {dtype} because no missing_value was provided\n\n"
                "Columns with dtype {dtype} require a missing_value.\n"
                "Please pass missing_value to Column() or use a different"
                " dtype.".format(dtype=dtype, name=name)
            ) from exc
        self.name = name
        self.doc = doc
        self.metadata = metadata
        self.currency_aware = currency_aware

    def __get__(self, instance, owner):
        """
        Produce a concrete BoundColumn object when accessed.

        We don't bind to datasets at class creation time so that subclasses of
        DataSets produce different BoundColumns.
        """
        return BoundColumn(
            dtype=self.dtype,
            missing_value=self.missing_value,
            dataset=owner,
            name=self.name,
            doc=self.doc,
            metadata=self.metadata,
            currency_conversion=None,
            currency_aware=self.currency_aware,
        )


class BoundColumn(LoadableTerm):
    """
    A column of data that's been concretely bound to a particular dataset.

    Attributes
    ----------
    dtype : numpy.dtype
        The dtype of data produced when this column is loaded.
    latest : ziplime.pipeline.LoadableTerm
        A :class:`~ziplime.pipeline.Filter`, :class:`~ziplime.pipeline.Factor`,
        or :class:`~ziplime.pipeline.Classifier` computing the most recently
        known value of this column on each date.
        See :class:`ziplime.pipeline.mixins.LatestMixin` for more details.
    dataset : ziplime.pipeline.data.DataSet
        The dataset to which this column is bound.
    name : str
        The name of this column.
    metadata : dict
        Extra metadata associated with this column.
    currency_aware : bool
        Whether or not this column produces currency-denominated data.

    Notes
    -----
    Instances of this class are dynamically created upon access to attributes
    of :class:`~ziplime.pipeline.data.DataSet`. For example,
    :attr:`~ziplime.pipeline.data.EquityPricing.close` is an instance of this
    class. Pipeline API users should never construct instances of this
    directly.
    """

    mask = AssetExists()
    window_safe = True

    def __new__(
            cls,
            dtype,
            missing_value,
            dataset,
            name,
            doc,
            metadata,
            currency_conversion,
            currency_aware,
    ):
        if currency_aware and dtype != float64_dtype:
            raise AssertionError(
                "The {} column on dataset {} cannot be constructed with "
                "currency_aware={}, dtype={}. Currency aware columns must "
                "have a float64 dtype.".format(
                    name,
                    dataset,
                    currency_aware,
                    dtype,
                )
            )

        return super(BoundColumn, cls).__new__(
            cls,
            domain=dataset.domain,
            dtype=dtype,
            missing_value=missing_value,
            dataset=dataset,
            name=name,
            ndim=dataset.ndim,
            doc=doc,
            metadata=metadata,
            currency_conversion=currency_conversion,
            currency_aware=currency_aware,
        )

    def _init(
            self,
            dataset,
            name,
            doc,
            metadata,
            currency_conversion,
            currency_aware,
            *args,
            **kwargs,
    ):
        self._dataset = dataset
        self._name = name
        self.__doc__ = doc
        self._metadata = metadata
        self._currency_conversion = currency_conversion
        self._currency_aware = currency_aware
        return super(BoundColumn, self)._init(*args, **kwargs)

    @classmethod
    def _static_identity(
            cls,
            dataset,
            name,
            doc,
            metadata,
            currency_conversion,
            currency_aware,
            *args,
            **kwargs,
    ):
        return (
            super(BoundColumn, cls)._static_identity(*args, **kwargs),
            dataset,
            name,
            doc,
            frozenset(sorted(metadata.items(), key=first)),
            currency_conversion,
            currency_aware,
        )

    def __lt__(self, other):
        msg = "Can't compare '{}' with '{}'. (Did you mean to use '.latest'?)"
        raise TypeError(msg.format(self.qualname, other.__class__.__name__))

    __gt__ = __le__ = __ge__ = __lt__

    def _replace(self, **kwargs):
        kw = dict(
            dtype=self.dtype,
            missing_value=self.missing_value,
            dataset=self._dataset,
            name=self._name,
            doc=self.__doc__,
            metadata=self._metadata,
            currency_conversion=self._currency_conversion,
            currency_aware=self._currency_aware,
        )
        kw.update(kwargs)

        return type(self)(**kw)

    def specialize(self, domain):
        """Specialize ``self`` to a concrete domain."""
        if domain == self.domain:
            return self

        return self._replace(dataset=self._dataset.specialize(domain))

    def unspecialize(self):
        """
        Unspecialize a column to its generic form.

        This is equivalent to ``column.specialize(GENERIC)``.
        """
        return self.specialize(GENERIC)

    def fx(self, currency: str | Currency):
        """
        Construct a currency-converted version of this column.

        Parameters
        ----------
        currency : str or ziplime.currency.Currency
            Currency into which to convert this column's data.

        Returns
        -------
        column : BoundColumn
            Column producing the same data as ``self``, but currency-converted
            into ``currency``.
        """
        conversion = self._currency_conversion

        if not self._currency_aware:
            raise TypeError(
                "The .fx() method cannot be called on {} because it does not "
                "produce currency-denominated data.".format(self.qualname)
            )
        elif conversion is not None and conversion.currency == currency:
            return self

        return self._replace(
            currency_conversion=CurrencyConversion(
                currency=currency,
                field=DEFAULT_FX_RATE,
            )
        )

    @property
    def currency_conversion(self):
        """Specification for currency conversions applied for this term."""
        return self._currency_conversion

    @property
    def currency_aware(self):
        """
        Whether or not this column produces currency-denominated data.
        """
        return self._currency_aware

    @property
    def dataset(self):
        """
        The dataset to which this column is bound.
        """
        return self._dataset

    @property
    def name(self):
        """
        The name of this column.
        """
        return self._name

    @property
    def metadata(self):
        """
        A copy of the metadata for this column.
        """
        return self._metadata.copy()

    @property
    def qualname(self):
        """The fully-qualified name of this column."""
        out = ".".join([self.dataset.qualname, self.name])
        conversion = self._currency_conversion
        if conversion is not None:
            out += ".fx({!r})".format(conversion.currency.code)
        return out

    @property
    def latest(self):
        dtype = self.dtype
        if dtype in Filter.ALLOWED_DTYPES:
            Latest = LatestFilter
        elif dtype in Classifier.ALLOWED_DTYPES:
            Latest = LatestClassifier
        else:
            assert dtype in Factor.ALLOWED_DTYPES, "Unknown dtype %s." % dtype
            Latest = LatestFactor

        return Latest(
            inputs=(self,),
            dtype=dtype,
            missing_value=self.missing_value,
            ndim=self.ndim,
        )

    def __repr__(self):
        return "{qualname}::{dtype}".format(
            qualname=self.qualname,
            dtype=self.dtype.name,
        )

    def graph_repr(self):
        """Short repr to use when rendering Pipeline graphs."""
        # Graphviz interprets `\l` as "divide label into lines, left-justified"
        return "BoundColumn:\\l  Dataset: {}\\l  Column: {}\\l".format(
            self.dataset.__name__, self.name
        )

    def recursive_repr(self):
        """Short repr used to render in recursive contexts."""
        return self.qualname


class DataSetMeta(type):
    """
    Metaclass for DataSets

    Supplies name and dataset information to Column attributes, and manages
    families of specialized dataset.
    """

    def __new__(metacls, name, bases, dict_):
        if len(bases) != 1:
            # Disallowing multiple inheritance makes it easier for us to
            # determine whether a given dataset is the root for its family of
            # specializations.
            raise TypeError("Multiple dataset inheritance is not supported.")

        # This marker is set in the class dictionary by `specialize` below.
        is_specialization = dict_.pop(IsSpecialization, False)

        newtype = super(DataSetMeta, metacls).__new__(metacls, name, bases, dict_)

        if not isinstance(newtype.domain, Domain):
            raise TypeError(
                "Expected a Domain for {}.domain, but got {} instead.".format(
                    newtype.__name__,
                    type(newtype.domain),
                )
            )

        # Collect all of the column names that we inherit from our parents.
        column_names = set().union(
            *(getattr(base, "_column_names", ()) for base in bases)
        )

        # Collect any new columns from this dataset.
        for maybe_colname, maybe_column in dict_.items():
            if isinstance(maybe_column, Column):
                # add column names defined on our class
                bound_column_descr = maybe_column.bind(maybe_colname)
                setattr(newtype, maybe_colname, bound_column_descr)
                column_names.add(maybe_colname)

        newtype._column_names = frozenset(column_names)

        if not is_specialization:
            # This is the new root of a family of specializations. Store the
            # memoized dictionary for family on this type.
            newtype._domain_specializations = WeakKeyDictionary(
                {
                    newtype.domain: newtype,
                }
            )

        return newtype

    def specialize(cls, domain: Domain):
        """
        Specialize a generic DataSet to a concrete domain.

        Parameters
        ----------
        domain : ziplime.pipeline.domain.Domain
            Domain to which we should generate a specialization.

        Returns
        -------
        specialized : type
            A new :class:`~ziplime.pipeline.data.DataSet` subclass with the
            same columns as ``self``, but specialized to ``domain``.
        """
        # We're already the specialization to this domain, so just return self.
        if domain == cls.domain:
            return cls

        try:
            return cls._domain_specializations[domain]
        except KeyError as exc:
            if not cls._can_create_new_specialization(domain):
                # This either means we're already a specialization and trying
                # to create a new specialization, or we're the generic version
                # of a root-specialized dataset, which we don't want to create
                # new specializations of.
                raise ValueError(
                    "Can't specialize {dataset} from {current} to new domain {new}.".format(
                        dataset=cls.__name__,
                        current=cls.domain,
                        new=domain,
                    )
                ) from exc
            new_type = cls._create_specialization(domain)
            cls._domain_specializations[domain] = new_type
            return new_type

    def unspecialize(cls):
        """
        Unspecialize a dataset to its generic form.

        This is equivalent to ``dataset.specialize(GENERIC)``.
        """
        return cls.specialize(GENERIC)

    def _can_create_new_specialization(cls, domain):
        # Always allow specializing to a generic domain.
        if domain is GENERIC:
            return True
        elif "_domain_specializations" in vars(cls):
            # This branch is True if we're the root of a family.
            # Allow specialization if we're generic.
            return cls.domain is GENERIC
        else:
            # If we're not the root of a family, we can't create any new
            # specializations.
            return False

    def _create_specialization(cls, domain):
        # These are all assertions because we should have handled these cases
        # already in specialize().
        assert isinstance(domain, Domain)
        assert (
                domain not in cls._domain_specializations
        ), "Domain specializations should be memoized!"
        if domain is not GENERIC:
            assert (
                    cls.domain is GENERIC
            ), "Can't specialize dataset with domain {} to domain {}.".format(
                cls.domain,
                domain,
            )

        # Create a new subclass of ``self`` with the given domain.
        # Mark that it's a specialization so that we know not to create a new
        # family for it.
        name = cls.__name__
        bases = (cls,)
        dict_ = {"domain": domain, IsSpecialization: True}
        out = type(name, bases, dict_)
        out.__module__ = cls.__module__
        return out

    @property
    def columns(cls):
        return frozenset(getattr(cls, colname) for colname in cls._column_names)

    @property
    def qualname(cls):
        if cls.domain is GENERIC:
            specialization_key = ""
        else:
            specialization_key = "<" + cls.domain.country_code + ">"

        return cls.__name__ + specialization_key

    # NOTE: We used to use `functools.total_ordering` to account for all of the
    #       other rich comparison methods, but it has issues in python 3 and
    #       this method is only used for test purposes, so for now we will just
    #       keep this in isolation. If we ever need any of the other comparison
    #       methods we will have to implement them individually.
    def __lt__(cls, other):
        return id(cls) < id(other)

    def __repr__(cls):
        return "<DataSet: %r, domain=%s>" % (cls.__name__, cls.domain)


class DataSet(object, metaclass=DataSetMeta):
    """
    Base class for Pipeline datasets.

    A :class:`DataSet` is defined by two parts:

    1. A collection of :class:`~ziplime.pipeline.data.Column` objects that
       describe the queryable attributes of the dataset.

    2. A :class:`~ziplime.pipeline.domain.Domain` describing the assets and
       calendar of the data represented by the :class:`DataSet`.

    To create a new Pipeline dataset, define a subclass of :class:`DataSet` and
    set one or more :class:`Column` objects as class-level attributes. Each
    column requires a ``np.dtype`` that describes the type of data that should
    be produced by a loader for the dataset. Integer columns must also provide
    a "missing value" to be used when no value is available for a given
    asset/date combination.

    By default, the domain of a dataset is the special singleton value,
    :data:`~ziplime.pipeline.domain.GENERIC`, which means that they can be used
    in a Pipeline running on **any** domain.

    In some cases, it may be preferable to restrict a dataset to only allow
    support a single domain. For example, a DataSet may describe data from a
    vendor that only covers the US. To restrict a dataset to a specific domain,
    define a `domain` attribute at class scope.

    You can also define a domain-specific version of a generic DataSet by
    calling its ``specialize`` method with the domain of interest.

    Examples
    --------
    The built-in EquityPricing dataset is defined as follows::

        class EquityPricing(DataSet):
            open = Column(float)
            high = Column(float)
            low = Column(float)
            close = Column(float)
            volume = Column(float)

    The built-in USEquityPricing dataset is a specialization of
    EquityPricing. It is defined as::

        from ziplime.pipeline.domain import US_EQUITIES
        USEquityPricing = EquityPricing.specialize(US_EQUITIES)

    Columns can have types other than float. A dataset containing assorted
    company metadata might be defined like this::

        class CompanyMetadata(DataSet):
            # Use float for semantically-numeric data, even if it's always
            # integral valued (see Notes section below). The default missing
            # value for floats is NaN.
            shares_outstanding = Column(float)

            # Use object for string columns. The default missing value for
            # object-dtype columns is None.
            ticker = Column(object)

            # Use integers for integer-valued categorical data like sector or
            # industry codes. Integer-dtype columns require an explicit missing
            # value.
            sector_code = Column(int, missing_value=-1)

            # Use bool for boolean-valued flags. Note that the default missing
            # value for bool-dtype columns is False.
            is_primary_share = Column(bool)

    Notes
    -----
    Because numpy has no native support for integers with missing values, users
    are strongly encouraged to use floats for any data that's semantically
    numeric. Doing so enables the use of `NaN` as a natural missing value,
    which has useful propagation semantics.
    """

    domain = GENERIC
    ndim = 2

    @classmethod
    def get_column(cls, name):
        """Look up a column by name.

        Parameters
        ----------
        name : str
            Name of the column to look up.

        Returns
        -------
        column : ziplime.pipeline.data.BoundColumn
            Column with the given name.

        Raises
        ------
        AttributeError
            If no column with the given name exists.
        """
        clsdict = vars(cls)
        try:
            maybe_column = clsdict[name]
            if not isinstance(maybe_column, _BoundColumnDescr):
                raise KeyError(name)
        except KeyError as exc:
            raise AttributeError(
                "{dset} has no column {colname!r}:\n\n"
                "Possible choices are:\n"
                "{choices}".format(
                    dset=cls.qualname,
                    colname=name,
                    choices=bulleted_list(
                        sorted(cls._column_names),
                        max_count=10,
                    ),
                )
            ) from exc

        # Resolve column descriptor into a BoundColumn.
        return maybe_column.__get__(None, cls)


# This attribute is set by DataSetMeta to mark that a class is the root of a
# family of datasets with diffent domains. We don't want that behavior for the
# base DataSet class, and we also don't want to accidentally use a shared
# version of this attribute if we fail to set this in a subclass somewhere.
del DataSet._domain_specializations


class DataSetFamilyLookupError(AttributeError):
    """Exception thrown when a column is accessed on a DataSetFamily
    instead of on the result of a slice.

    Parameters
    ----------
    family_name : str
        The name of the DataSetFamily on which the access occurred.
    column_name : str
        The name of the column accessed.
    """

    def __init__(self, family_name, column_name):
        self.family_name = family_name
        self.column_name = column_name

    def __str__(self):
        # NOTE: when ``aggregate`` is added, remember to update this message
        return dedent(
            """\
            Attempted to access column {c} from DataSetFamily {d}:

            To work with dataset families, you must first select a
            slice using the ``slice`` method:

                {d}.slice(...).{c}
            """.format(
                c=self.column_name, d=self.family_name
            )
        )


class _DataSetFamilyColumn:
    """Descriptor used to raise a helpful error when a column is accessed on a
    DataSetFamily instead of on the result of a slice.

    Parameters
    ----------
    column_names : str
        The name of the column.
    """

    def __init__(self, column_name):
        self.column_name = column_name

    def __get__(self, instance, owner):
        raise DataSetFamilyLookupError(
            owner.__name__,
            self.column_name,
        )


class DataSetFamilyMeta(abc.ABCMeta):
    def __new__(cls, name, bases, dict_):
        columns = {}
        for k, v in dict_.items():
            if isinstance(v, Column):
                # capture all the columns off the DataSetFamily class
                # and replace them with a descriptor that will raise a helpful
                # error message. The columns will get added to the BaseSlice
                # for this type.
                columns[k] = v
                dict_[k] = _DataSetFamilyColumn(k)

        is_abstract = dict_.pop("_abstract", False)

        self = super(DataSetFamilyMeta, cls).__new__(
            cls,
            name,
            bases,
            dict_,
        )

        if not is_abstract:
            self.extra_dims = extra_dims = OrderedDict(
                [(k, frozenset(v)) for k, v in OrderedDict(self.extra_dims).items()]
            )
            if not extra_dims:
                raise ValueError(
                    "DataSetFamily must be defined with non-empty"
                    " extra_dims, or with `_abstract = True`",
                )

            class BaseSlice(self._SliceType):
                dataset_family = self

                ndim = self.slice_ndim
                domain = self.domain

                locals().update(columns)

            BaseSlice.__name__ = "%sBaseSlice" % self.__name__
            self._SliceType = BaseSlice

        # each type gets a unique cache
        self._slice_cache = {}
        return self

    def __repr__(self):
        return "<DataSetFamily: %r, extra_dims=%r>" % (
            self.__name__,
            list(self.extra_dims),
        )


class DataSetFamilySlice(DataSet):
    """Marker type for slices of a
    :class:`ziplime.pipeline.data.dataset.DataSetFamily` objects
    """


# XXX: This docstring was mostly written when the abstraction here was
# "MultiDimensionalDataSet". It probably needs some rewriting.
class DataSetFamily(metaclass=DataSetFamilyMeta):
    """
    Base class for Pipeline dataset families.

    Dataset families are used to represent data where the unique identifier for
    a row requires more than just asset and date coordinates. A
    :class:`DataSetFamily` can also be thought of as a collection of
    :class:`~ziplime.pipeline.data.DataSet` objects, each of which has the same
    columns, domain, and ndim.

    :class:`DataSetFamily` objects are defined with one or more
    :class:`~ziplime.pipeline.data.Column` objects, plus one additional field:
    ``extra_dims``.

    The ``extra_dims`` field defines coordinates other than asset and date that
    must be fixed to produce a logical timeseries. The column objects determine
    columns that will be shared by slices of the family.

    ``extra_dims`` are represented as an ordered dictionary where the keys are
    the dimension name, and the values are a set of unique values along that
    dimension.

    To work with a :class:`DataSetFamily` in a pipeline expression, one must
    choose a specific value for each of the extra dimensions using the
    :meth:`~ziplime.pipeline.data.DataSetFamily.slice` method.
    For example, given a :class:`DataSetFamily`:

    .. code-block:: python

       class SomeDataSet(DataSetFamily):
           extra_dims = [
               ('dimension_0', {'a', 'b', 'c'}),
               ('dimension_1', {'d', 'e', 'f'}),
           ]

           column_0 = Column(float)
           column_1 = Column(bool)

    This dataset might represent a table with the following columns:

    ::

      sid :: int64
      asof_date :: datetime64[ns]
      timestamp :: datetime64[ns]
      dimension_0 :: str
      dimension_1 :: str
      column_0 :: float64
      column_1 :: bool

    Here we see the implicit ``sid``, ``asof_date`` and ``timestamp`` columns
    as well as the extra dimensions columns.

    This :class:`DataSetFamily` can be converted to a regular :class:`DataSet`
    with:

    .. code-block:: python

       DataSetSlice = SomeDataSet.slice(dimension_0='a', dimension_1='e')

    This sliced dataset represents the rows from the higher dimensional dataset
    where ``(dimension_0 == 'a') & (dimension_1 == 'e')``.
    """

    _abstract = True  # Removed by metaclass

    domain = GENERIC
    slice_ndim = 2

    _SliceType = DataSetFamilySlice

    @type.__call__
    class extra_dims:
        """OrderedDict[str, frozenset] of dimension name -> unique values

        May be defined on subclasses as an iterable of pairs: the
        metaclass converts this attribute to an OrderedDict.
        """

        __isabstractmethod__ = True

        def __get__(self, instance, owner):
            return []

    @classmethod
    def _canonical_key(cls, args, kwargs):
        extra_dims = cls.extra_dims
        dimensions_set = set(extra_dims)
        if not set(kwargs) <= dimensions_set:
            extra = sorted(set(kwargs) - dimensions_set)
            raise TypeError(
                "%s does not have the following %s: %s\n"
                "Valid dimensions are: %s"
                % (
                    cls.__name__,
                    s("dimension", extra),
                    ", ".join(extra),
                    ", ".join(extra_dims),
                ),
            )

        if len(args) > len(extra_dims):
            raise TypeError(
                "%s has %d extra %s but %d %s given"
                % (
                    cls.__name__,
                    len(extra_dims),
                    s("dimension", extra_dims),
                    len(args),
                    plural("was", "were", args),
                ),
            )

        missing = object()
        coords = OrderedDict(zip(extra_dims, repeat(missing)))
        to_add = dict(zip(extra_dims, args))
        coords.update(to_add)
        added = set(to_add)

        for key, value in kwargs.items():
            if key in added:
                raise TypeError(
                    "%s got multiple values for dimension %r"
                    % (
                        cls.__name__,
                        coords,
                    ),
                )
            coords[key] = value
            added.add(key)

        missing = {k for k, v in coords.items() if v is missing}
        if missing:
            missing = sorted(missing)
            raise TypeError(
                "no coordinate provided to %s for the following %s: %s"
                % (
                    cls.__name__,
                    s("dimension", missing),
                    ", ".join(missing),
                ),
            )

        # validate that all of the provided values exist along their given
        # dimensions
        for key, value in coords.items():
            if value not in cls.extra_dims[key]:
                raise ValueError(
                    "%r is not a value along the %s dimension of %s"
                    % (
                        value,
                        key,
                        cls.__name__,
                    ),
                )

        return coords, tuple(coords.items())

    @classmethod
    def _make_dataset(cls, coords):
        """Construct a new dataset given the coordinates."""

        class Slice(cls._SliceType):
            extra_coords = coords

        Slice.__name__ = "%s.slice(%s)" % (
            cls.__name__,
            ", ".join("%s=%r" % item for item in coords.items()),
        )
        return Slice

    @classmethod
    def slice(cls, *args, **kwargs):
        """Take a slice of a DataSetFamily to produce a dataset
        indexed by asset and date.

        Parameters
        ----------
        *args
        **kwargs
            The coordinates to fix along each extra dimension.

        Returns
        -------
        dataset : DataSet
            A regular pipeline dataset indexed by asset and date.

        Notes
        -----
        The extra dimensions coords used to produce the result are available
        under the ``extra_coords`` attribute.
        """
        coords, hash_key = cls._canonical_key(args, kwargs)
        try:
            return cls._slice_cache[hash_key]
        except KeyError:
            pass

        Slice = cls._make_dataset(coords)
        cls._slice_cache[hash_key] = Slice
        return Slice


CurrencyConversion = namedtuple(
    "CurrencyConversion",
    ["currency", "field"],
)



import datetime
from functools import lru_cache

import polars as pl
import uuid

from exchange_calendars import ExchangeCalendar

from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.constants.period import Period
from ziplime.data.domain.data_bundle import DataBundle

from ziplime.domain.position import Position
from ziplime.domain.portfolio import Portfolio
from ziplime.domain.account import Account
from ziplime.finance.commission import EquityCommissionModel, FutureCommissionModel, CommissionModel
from ziplime.finance.domain.order import Order
from ziplime.finance.slippage.slippage_model import SlippageModel
from ziplime.exchanges.exchange import Exchange
from ziplime.gens.domain.trading_clock import TradingClock


class SimulationExchange(Exchange):

    def __init__(self,
                 name: str,
                 country_code: str,
                 trading_calendar: ExchangeCalendar,
                 clock: TradingClock,
                 cash_balance: float,
                 equity_slippage: SlippageModel,
                 future_slippage: SlippageModel,
                 equity_commission: EquityCommissionModel,
                 future_commission: FutureCommissionModel,
                 data_bundle: DataBundle = None,
                 extra_data_sources: list = None
                 ):
        super().__init__(name=name,
                         canonical_name=name,
                         clock=clock,
                         data_bundle=data_bundle,
                         country_code=country_code,
                         trading_calendar=trading_calendar)
        self.slippage_models = {
            Equity: equity_slippage,
            FuturesContract: future_slippage,
        }
        self.commission_models = {
            Equity: equity_commission,
            FuturesContract: future_commission,
        }
        self.cash_balance = cash_balance

    def get_start_cash_balance(self) -> float:
        return self.cash_balance

    def get_current_cash_balance(self) -> float:
        return self.cash_balance

    def get_commission_model(self, asset: Asset) -> CommissionModel:
        return self.commission_models[type(asset)]

    def get_slippage_model(self, asset: Asset) -> SlippageModel:
        return self.slippage_models[type(asset)]

    async def submit_order(self, order: Order):
        order.id = uuid.uuid4().hex
        return order

    def get_positions(self) -> dict[Asset, Position]:
        pass

    def get_portfolio(self) -> Portfolio:
        pass

    def get_account(self) -> Account:
        pass

    def get_time_skew(self):
        pass

    def order(self, asset, amount, style):
        pass

    def is_alive(self):
        pass

    def get_orders(self) -> dict[str, Order]:
        pass

    async def get_transactions(self, orders: dict[Asset, dict[str, Order]],
                               current_dt: datetime.datetime):
        """
        Creates a list of transactions based on the current open orders,
        slippage model, and commission model.

        Parameters
        ----------
        bar_data: ziplime._protocol.BarData

        Notes
        -----
        This method book-keeps the blotter's open_orders dictionary, so that
         it is accurate by the time we're done processing open orders.

        Returns
        -------
        transactions_list: List
            transactions_list: list of transactions resulting from the current
            open orders.  If there were no open orders, an empty list is
            returned.

        commissions_list: List
            commissions_list: list of commissions resulting from filling the
            open orders.  A commission is an object with "asset" and "cost"
            parameters.

        closed_orders: List
            closed_orders: list of all the orders that have filled.
        """

        closed_orders = []
        transactions = []
        commissions = []

        for asset, asset_orders in orders.items():
            slippage = self.get_slippage_model(asset=asset)

            for order, txn in slippage.simulate(exchange=self, assets=frozenset({asset}),
                                                orders_for_asset=asset_orders.values(),
                                                current_dt=current_dt
                                                ):
                commission = self.get_commission_model(asset=asset)
                additional_commission = commission.calculate(order=order, transaction=txn)

                if additional_commission > 0:
                    commissions.append(
                        {
                            "asset": order.asset,
                            "order": order,
                            "cost": additional_commission,
                        }
                    )

                order.filled += txn.amount
                order.commission += additional_commission

                order.dt = txn.dt

                transactions.append(txn)

                if not order.open:
                    closed_orders.append(order)

        return transactions, commissions, closed_orders

    def get_orders_by_ids(self, order_ids: list[str]):
        pass

    def get_transactions_by_order_ids(self, order_ids: list[str]):
        pass

    def cancel_order(self, order_param):
        pass

    def get_last_traded_dt(self, asset):
        pass

    def get_spot_value(self, assets: frozenset[Asset], fields: frozenset[str], dt: datetime.datetime,
                       data_frequency: datetime.timedelta) -> pl.DataFrame:
        pass

    def get_realtime_bars(self, assets, frequency):
        pass

    async def get_scalar_asset_spot_value(self, asset: Asset, field: str, dt: datetime.datetime,
                                          frequency: datetime.timedelta):
        return self.get_data_by_limit(
            fields=frozenset({field}),
            limit=1,
            end_date=dt,
            frequency=frequency,
            assets=frozenset({asset}),
            include_end_date=True,
        )

    def get_scalar_asset_spot_value_sync(self, asset: Asset, field: str, dt: datetime.datetime,
                                         frequency: datetime.timedelta):

        return  self.get_data_by_limit(
            fields=frozenset({field}),
            limit=1,
            end_date=dt,
            frequency=frequency,
            assets=frozenset({asset}),
            include_end_date=True,
        )

    # @lru_cache(maxsize=100)
    def current(self, assets: frozenset[Asset], fields: frozenset[str], dt: datetime.datetime):
        data = {}
        # print(f"Getting current: {assets}, fields={fields}, dt={dt}")
        # TODO: check this, uncomment adjust_minutes
        # if not self._adjust_minutes:
        # return self.data_bundle.get_spot_value(
        #     assets=assets,
        #     fields=fields,
        #     dt=dt,
        #     frequency=self.data_bundle.frequency
        # )

        return self.get_data_by_limit(
            fields=fields,
            limit=1,
            end_date=dt,
            frequency=self.data_bundle.frequency,
            assets=assets,
            include_end_date=True,
        )
        # return df_raw

    @lru_cache(maxsize=100)
    def get_data_by_period(self,
                           fields: frozenset[str],
                          start_date: datetime.datetime,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          source: str
                          ) -> pl.DataFrame:
        return self.data_bundle.get_data_by_limit(fields=fields,
                                                  limit=limit,
                                                  end_date=end_date,
                                                  frequency=frequency,
                                                  assets=assets,
                                                  include_end_date=include_end_date,
                                                  )

    @lru_cache(maxsize=100)
    def get_data_by_limit(self, fields: frozenset[str],
                          limit: int,
                          end_date: datetime.datetime,
                          frequency: datetime.timedelta | Period,
                          assets: frozenset[Asset],
                          include_end_date: bool,
                          ) -> pl.DataFrame:
        return self.data_bundle.get_data_by_limit(fields=fields,
                                                  limit=limit,
                                                  end_date=end_date,
                                                  frequency=frequency,
                                                  assets=assets,
                                                  include_end_date=include_end_date,
                                                  )



import pandas as pd
from exchange_calendars import get_calendar as ec_get_calendar  # get_calendar,


def get_calendar(*args, **kwargs):
    if args[0] in ["us_futures", "CMES", "XNYS", "NYSE"]:
        return ec_get_calendar(*args, side="right", start=pd.Timestamp("1886-01-01"))
    return ec_get_calendar(*args, side="right")



from sqlalchemy import PrimaryKeyConstraint
from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import EquityFK, StringIndexed, DateIndexed
from ziplime.core.db.base_model import BaseModel


class EquitySupplementaryMapping(BaseModel):
    __tablename__ = "equity_supplementary_mappings"

    sid: Mapped[EquityFK]
    field: Mapped[StringIndexed]
    start_date: Mapped[DateIndexed]
    end_date: Mapped[DateIndexed]
    value: Mapped[str]

    __table_args__ = (
        PrimaryKeyConstraint('sid', 'field', "start_date"),
    )



import datetime
from exchange_calendars import ExchangeCalendar
import polars as pl


class SimulationParameters:
    def __init__(
            self,
            start_date: datetime.datetime,
            end_date: datetime.datetime,
            trading_calendar: ExchangeCalendar,
            emission_rate: datetime.timedelta,
    ):

        if start_date >= end_date:
            raise ValueError("Period start falls after period end.")
        if start_date >= trading_calendar.last_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError("Period start falls after the last known trading day.")
        if end_date <= trading_calendar.first_session.replace(tzinfo=trading_calendar.tz):
            raise ValueError("Period end falls before the first known trading day.")

        # chop off any minutes or hours on the given start and end dates,
        # as we only support session labels here (and we represent session
        # labels as midnight UTC).

        self.start_date = start_date
        self.end_date = end_date

        self.start_session = start_date.date()
        self.end_session = end_date.date()

        self.emission_rate = emission_rate

        self.trading_calendar = trading_calendar
        if not trading_calendar.is_session(self.start_session):
            # if the start date is not a valid session in this calendar,
            # push it forward to the first valid session
            self.start_session = trading_calendar.minute_to_session(
                self.start_session
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        if not trading_calendar.is_session(self.end_session):
            # if the end date is not a valid session in this calendar,
            # pull it backward to the last valid session before the given
            # end date.
            self.end_session = trading_calendar.minute_to_session(
                self.end_session, direction="previous"
            ).tz_localize(self.trading_calendar.tz).to_pydatetime().date()

        self.first_open = trading_calendar.session_first_minute(
            self.start_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()
        self.last_close = trading_calendar.session_close(
            self.end_session
        ).tz_convert(self.trading_calendar.tz).to_pydatetime()

        self.sessions = pl.Series(self.trading_calendar.sessions_in_range(
            self.start_session, self.end_session)
        ).dt.date()

        self.market_closes = pl.Series(
            self.trading_calendar.schedule.loc[self.sessions, "close"].dt.tz_convert(
                self.trading_calendar.tz))
        self.market_opens = pl.Series(
            self.trading_calendar.first_minutes.loc[self.sessions].dt.tz_convert(
                self.trading_calendar.tz))

        self.before_trading_start_minutes = self.market_opens - datetime.timedelta(minutes=46)



import asyncio
import datetime
import logging
from functools import partial

import structlog
from lime_trader.models.page import PageRequest
from lime_trader.utils.pagination import iterate_pages_async

from ziplime.assets.entities.asset import Asset
from ziplime.domain.bar_data import BarData

from ziplime.domain.portfolio import Portfolio as ZpPortfolio
from ziplime.domain.position import Position as ZpPosition
from ziplime.domain.account import Account as ZpAccount

from lime_trader.models.accounts import AccountDetails, TradeSide
from lime_trader.models.market import Period
from lime_trader.models.trading import Order as LimeTraderOrder, OrderSide, OrderDetails, \
    OrderStatus as LimeTraderOrderStatus, OrderType, TimeInForce
from ziplime.finance.execution import (MarketOrder,
                                       LimitOrder,
                                       )

from ziplime.finance.commission import CommissionModel
from ziplime.finance.domain.order import Order
from ziplime.finance.domain.order_status import OrderStatus
from ziplime.finance.domain.transaction import Transaction
import pandas as pd
import numpy as np
import uuid

from ziplime.finance.slippage.slippage_model import SlippageModel
from ziplime.exchanges.exchange import Exchange


class TradingSignalExchange(Exchange):

    def __init__(self, name: str):
        super().__init__(name=name)
        self._logger = structlog.get_logger(__name__)

    def get_positions(self) -> dict[Asset, ZpPosition]:
        return {}

    def get_portfolio(self) -> ZpPortfolio:
        return {}

    def get_account(self) -> ZpAccount:
        account = self.get_account_balance(account_number=self._account_id)
        z_account = ZpAccount()
        z_account.buying_power = float(account.cash)
        z_account.total_position_value = float(account.position_market_value)
        return z_account

    def get_account_balance(self, account_number: str) -> AccountDetails:
        acc = next(filter(lambda x: x.account_number == account_number, self._lime_sdk_client.account.get_balances()),
                   None)
        if acc is None:
            raise Exception(f"Invalid account number {account_number}. Not found.")
        return acc

    def get_time_skew(self) -> pd.Timedelta:
        return pd.Timedelta('0 sec')  # TODO: use clock API

    def is_alive(self) -> bool:
        try:
            self._lime_sdk_client.account.get_balances()
            return True
        except Exception as _:
            return False

    def _order2zp(self, order: OrderDetails, asset: Asset) -> Order | None:

        match order.order_status:
            case LimeTraderOrderStatus.CANCELED:
                order_status = OrderStatus.CANCELLED
            case LimeTraderOrderStatus.REJECTED:
                order_status = OrderStatus.REJECTED
            case LimeTraderOrderStatus.SUSPENDED:
                order_status = OrderStatus.HELD
            # case LimeTraderOrderStatus.REPLACED:
            #     zp_order.status = OrderStatus.REPLACED
            # case LimeTraderOrderStatus.PENDING_CANCEL:
            #     zp_order.status = OrderStatus.PENDING_CANCEL
            # case LimeTraderOrderStatus.DONE_FOR_DAY:
            #     zp_order.status = OrderStatus.DONE_FOR_DAY
            case LimeTraderOrderStatus.NEW:
                order_status = OrderStatus.OPEN
            case LimeTraderOrderStatus.PENDING_NEW:
                order_status = OrderStatus.OPEN
            case LimeTraderOrderStatus.PARTIALLY_FILLED:
                order_status = OrderStatus.OPEN
            case LimeTraderOrderStatus.FILLED:
                order_status = OrderStatus.FILLED
            case _:
                raise Exception(f"Unknown order status: {order.order_status}")
        match order.order_type:
            case OrderType.MARKET:
                execution_style = MarketOrder()
            case OrderType.LIMIT:
                execution_style = LimitOrder(limit_price=float(order.price))
            case _:
                raise Exception(f"Unknown order type {order.order_type}")

        order_details = Order(
            id=order.client_order_id,
            asset=asset,
            amount=int(order.quantity) if order.order_side == OrderSide.BUY else -int(order.quantity),
            filled=int(order.executed_quantity),
            dt=order.executed_timestamp,
            commission=0.0,
            execution_style=execution_style,
            status=order_status,
            exchange_order_id=order.order_id
        )

        return order_details

    def get_orders(self) -> dict[str, Order]:
        # return {}
        current_active_orders = self._lime_sdk_client.trading.get_active_orders(
            account_number=self._account_id)

        for order in current_active_orders:
            self._tracked_orders[order.client_order_id] = order

        result = {}
        for o in self._tracked_orders.values():
            ziplime_order = self._order2zp(order=o)
            if ziplime_order is None:
                continue
            result[o.client_order_id] = ziplime_order

        return result

    def get_orders_by_ids(self, order_ids: list[str]) -> list[OrderDetails]:
        result = []
        for order_id in order_ids:
            order = self._lime_sdk_client.trading.get_order_details_by_client_order_id(client_order_id=order_id)
            result.append(order)
        return result

    async def get_transactions(self, orders: dict[Asset, dict[str, Order]], bar_data: BarData):
        closed_orders = []
        transactions = []
        commissions = []

        all_orders = []
        for asset, asset_orders_dict in orders.items():
            asset_orders = list(asset_orders_dict.values())
            all_orders.extend(asset_orders)
        if not all_orders:
            return transactions, commissions, closed_orders
        assets_from_orders = {order.asset.get_symbol_by_exchange(self.name): order.asset for order in all_orders}
        start_date_for_transactions = min(o.dt for o in all_orders)

        orders = [self._lime_sdk_client.trading.get_order_details(order_id=order.exchange_order_id) for
                  order in all_orders]
        order_details = await asyncio.gather(*orders)

        for order_sdk_raw, order in zip(order_details, all_orders):
            order_sdk = self._order2zp(order=order_sdk_raw, asset=order.asset)
            if order_sdk_raw.executed_timestamp is None:
                continue
            if not order_sdk.open:
                closed_orders.append(order)
        async for transaction_page in iterate_pages_async(start_page=PageRequest(page=1, size=20),
                                              func=partial(self._lime_sdk_client.account.get_trades,
                                                           account_number=self._account_id,
                                                           date=bar_data.current_dt.date())):
            # async for transaction_page in self._lime_sdk_client.account.iterate_trades(
            #         account_number=self._account_id,
            #         start_page=PageRequest(page=1, size=20),
            #         date=start_date_for_transactions.date()):

            for transaction_sdk in transaction_page.data:
                asset = assets_from_orders[transaction_sdk.symbol]
                if asset is None:
                    continue

                # total_commissions = sum(
                #     fee.amount for fee in transaction_sdk.fees
                # )
                total_commissions = 0.0

                if transaction_sdk.trade_id in self.processed_transaction_ids:
                    continue

                tx = Transaction(
                    asset=asset,
                    amount=transaction_sdk.quantity if transaction_sdk.side == TradeSide.BUY else -transaction_sdk.quantity,
                    dt=transaction_sdk.timestamp,
                    price=float(transaction_sdk.price),
                    order_id=None,
                    commission=total_commissions,  # TODO: how to get commission
                )
                transactions.append(tx)
                self.processed_transaction_ids.add(transaction_sdk.trade_id)
                commissions.append(
                    {
                        "asset": asset,
                        # "order": order,
                        "cost": total_commissions,
                    }
                )
        return transactions, commissions, closed_orders

        #
        #     for order, txn in slippage.simulate(data=bar_data, assets={asset},
        #                                         orders_for_asset=asset_orders.values()):
        #         commission = self.get_commission_model(asset=asset)
        #         additional_commission = commission.calculate(order, txn)
        #
        #         if additional_commission > 0:
        #
        #
        #         order.filled += txn.amount
        #         order.commission += additional_commission
        #
        #         order.dt = txn.dt
        #
        #         transactions.append(txn)
        #
        #         if not order.open:
        #             closed_orders.append(order)
        #
        # for asset, asset_orders_dict in orders.items():
        #     asset_orders = list(asset_orders_dict.values())
        #     orders = [self._lime_sdk_client.trading.get_order_details_by_client_order_id(client_order_id=order.id) for
        #               order in asset_orders]
        #     order_details = await asyncio.gather(*orders)
        #     for order_sdk_raw, order in zip(order_details, asset_orders):
        #         order_sdk = self._order2zp(order=order_sdk_raw, asset=asset)
        #         if order_sdk_raw.executed_timestamp is None:
        #             continue
        #         tx = Transaction(
        #             asset=order.asset,
        #             amount=int(order_sdk_raw.executed_quantity),
        #             dt=order_sdk_raw.executed_timestamp,
        #             price=float(order_sdk_raw.price),
        #             order_id=order_sdk_raw.client_order_id,
        #             commission=order_sdk_raw,
        #         )
        #         if not order_sdk.open:
        #             closed_orders.append(order_sdk)
        #         results[order_sdk_raw.client_order_id] = tx
        return results

        # raise NotImplementedError("Use get_transactions_by_order_ids method.")

    # def get_transactions_by_order_ids(self, order_ids: list[str]):
    #     results = {}
    #
    #     for order_id in order_ids:
    #         order = self._lime_sdk_client.trading.get_order_details_by_client_order_id(client_order_id=order_id)
    #         # self._lime_sdk_client.account.iterate_trades(account_number=self._account_id,
    #         #                                          date=
    #         #
    #         #                                          ):
    #         if order.executed_timestamp is None:
    #             continue
    #         try:
    #             asset = symbol_lookup(order.symbol)
    #         except SymbolNotFound:
    #             continue
    #         tx = Transaction(
    #             asset=asset,
    #             amount=int(order.executed_quantity),
    #             dt=order.executed_timestamp,
    #             price=float(order.price),
    #             order_id=order.client_order_id,
    #             commission=0.0,
    #         )
    #         results[order.client_order_id] = tx
    #     return results

    def cancel_order(self, zp_order_id: str) -> None:
        try:
            order = self._lime_sdk_client.trading.get_order_details_by_client_order_id(order_id=zp_order_id)
            self._lime_sdk_client.trading.cancel_order(order_id=order.order_id)
        except Exception as e:
            self._logger.error(e)
            return

    def get_last_traded_dt(self, asset) -> datetime.datetime:
        quote = self._lime_sdk_client.market.get_current_quote(asset.symbol)
        return quote.date

    def get_spot_value(self, assets, field, dt, data_frequency):
        assert (field in (
            'open', 'high', 'low', 'close', 'volume', 'price'))
        assets_is_scalar = not isinstance(assets, (list, set, tuple))
        if assets_is_scalar:
            symbols = [assets.symbol]
        else:
            symbols = [asset.symbol for asset in assets]
        if field in ('price',):
            quotes = self._lime_sdk_client.market.get_current_quotes(symbols=symbols)
            if assets_is_scalar:
                if field == 'price':
                    if len(quotes) == 0:
                        return np.nan
                    return float(quotes[-1].last)
                else:
                    if len(quotes) == 0:
                        return pd.NaT
                    return quotes[-1].last_timestamp
            else:
                return [
                    float(quote.last) if field == 'price' else quote.last_timestamp
                    for quote in quotes
                ]

        bars_list = self._lime_sdk_client.market.get_quotes_history(symbol=symbols[0], period=Period.MINUTE,
                                                                    from_date=dt,
                                                                    to_date=dt)
        if assets_is_scalar:
            if len(bars_list) == 0:
                return np.nan
            return getattr(bars_list[0], field)
        bars_map = {a.symbol: a for a in bars_list}
        return [
            getattr(bars_map[symbol], field)
            for symbol in symbols
        ]

    def get_realtime_bars(self, assets, data_frequency):
        # TODO: cache the result. The caller
        # (DataPortalLive#get_history_window) makes use of only one
        # column at a time.
        assets_is_scalar = not isinstance(assets, (list, set, tuple))
        is_daily = 'd' in data_frequency  # 'daily' or '1d'
        if assets_is_scalar:
            symbols = [assets.symbol]
        else:
            symbols = [asset.symbol for asset in assets]
        timeframe = Period.DAY if is_daily else Period.MINUTE
        if not symbols:
            return []
        dfs = []
        to_date = datetime.datetime.now(tz=datetime.timezone.utc)
        from_date = to_date - datetime.timedelta(days=1)
        for asset in assets if not assets_is_scalar else [assets]:

            bars_list = self._lime_sdk_client.market.get_quotes_history(symbol=asset.symbol, period=timeframe,
                                                                        from_date=from_date,
                                                                        to_date=to_date
                                                                        )
            bars_map = {a.symbol: a for a in bars_list}

            symbol = asset.symbol
            df = bars_map[symbol].df.copy()
            if df.index.tz is None:
                df.index = df.index.tz_localize(
                    'utc').tz_convert('America/New_York')
            df.columns = pd.MultiIndex.from_product([[asset, ], df.columns])
            dfs.append(df)
        return pd.concat(dfs, axis=1)

    async def submit_order(self, order: Order):
        symbol = order.asset.get_symbol_by_exchange(self.name)
        qty = order.amount if order.amount > 0 else -order.amount
        side = OrderSide.BUY if order.amount > 0 else OrderSide.SELL

        if isinstance(order.execution_style, MarketOrder):
            order_type = OrderType.MARKET
        elif isinstance(order.execution_style, LimitOrder):
            order_type = OrderType.LIMIT
        else:
            raise Exception(f"Unsupported order type: {order.execution_style}.")

        new_order_id = uuid.uuid4().hex
        sdk_order = LimeTraderOrder(
            symbol=symbol,
            quantity=float(qty),
            side=side,
            order_type=order_type,
            time_in_force=TimeInForce.DAY,
            price=order.limit,
            client_order_id=new_order_id,
            account_number=self._account_id
        )
        validated_order = await self._lime_sdk_client.trading.validate_order(order=sdk_order)
        if not validated_order.is_valid:
            raise Exception("Validation failed for order")

        submitted_order = await self._lime_sdk_client.trading.place_order(order=sdk_order)
        # await asyncio.sleep(5)  # TODO: check why order is not immediately available
        order_details_sdk = await self._lime_sdk_client.trading.get_order_details(
            order_id=submitted_order.order_id
        )
        order_details = self._order2zp(order=order_details_sdk, asset=order.asset)

        return order_details

    def get_commission_model(self, asset: Asset) -> CommissionModel:
        pass

    def get_slippage_model(self, asset: Asset) -> SlippageModel:
        pass



#
# Copyright 2013 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

import pandas as pd
import structlog

log = structlog.get_logger(__name__)


def get_benchmark_returns_from_file(filelike):
    """Get a Series of benchmark returns from a file

    Parameters
    ----------
    filelike : str or file-like object
        Path to the benchmark file.
        expected csv file format:
        date,return
        2020-01-02 00:00:00+00:00,0.01
        2020-01-03 00:00:00+00:00,-0.02

    """
    log.info("Reading benchmark returns from %s", filelike)

    df = pd.read_csv(
        filelike,
        index_col=["date"],
        parse_dates=["date"],
    )
    if df.index.tz is not None:
        df = df.tz_localize(None)

    if "return" not in df.columns:
        raise ValueError(
            "The column 'return' not found in the "
            "benchmark file \n"
            "Expected benchmark file format :\n"
            "date, return\n"
            "2020-01-02 00:00:00+00:00,0.01\n"
            "2020-01-03 00:00:00+00:00,-0.02\n"
        )

    return df["return"].sort_index()



import datetime
from dataclasses import dataclass


@dataclass(frozen=True)
class CurrencySymbolMapping:
    symbol: str
    start_date: datetime.date
    end_date: datetime.date
    exchange_name: str



"""Pipeline hooks for tracking and displaying progress.
"""

from collections import namedtuple
import time

from ziplime.utils.compat import contextmanager, escape_html
from ziplime.utils.string_formatting import bulleted_list

from .iface import PipelineHooks


class ProgressHooks(PipelineHooks):
    """
    Hooks implementation for displaying progress.

    Parameters
    ----------
    publisher_factory : callable
        Function producing a new object with a ``publish()`` method that takes
        a ``ProgressModel`` and publishes progress to a consumer.
    """

    def __init__(self, publisher_factory):
        self._publisher_factory = publisher_factory
        self._reset_transient_state()

    def _reset_transient_state(self):
        self._start_date = None
        self._end_date = None
        self._model = None
        self._publisher = None

    @classmethod
    def with_widget_publisher(cls):
        """
        Construct a ProgressHooks that publishes to Jupyter via
        ``IPython.display``.
        """
        return cls(publisher_factory=IPythonWidgetProgressPublisher)

    @classmethod
    def with_static_publisher(cls, publisher):
        """Construct a ProgressHooks that uses an already-constructed publisher."""
        return cls(publisher_factory=lambda: publisher)

    def _publish(self):
        self._publisher.publish(self._model)

    @contextmanager
    def running_pipeline(self, pipeline, start_date, end_date):
        self._start_date = start_date
        self._end_date = end_date

        try:
            yield
        except Exception:
            if self._model is None:
                # This will only happen if an error happens in the Pipeline
                # Engine beteween entering `running_pipeline` and the first
                # `computing_chunk` call. If that happens, just propagate the
                # exception.
                raise
            self._model.finish(success=False)
            self._publish()
            raise
        else:
            self._model.finish(success=True)
            self._publish()
        finally:
            self._reset_transient_state()

    @contextmanager
    def computing_chunk(self, terms, start_date, end_date):
        # Set up model on first compute_chunk call.
        if self._model is None:
            self._publisher = self._publisher_factory()
            self._model = ProgressModel(
                start_date=self._start_date,
                end_date=self._end_date,
            )

        try:
            self._model.start_chunk(terms, start_date, end_date)
            self._publish()
            yield
        finally:
            self._model.finish_chunk(terms, start_date, end_date)
            self._publish()

    @contextmanager
    def loading_terms(self, terms):
        try:
            self._model.start_load_terms(terms)
            self._publish()
            yield
        finally:
            self._model.finish_load_terms(terms)
            self._publish()

    @contextmanager
    def computing_term(self, term):
        try:
            self._model.start_compute_term(term)
            self._publish()
            yield
        finally:
            self._model.finish_compute_term(term)
            self._publish()


class ProgressModel:
    """
    Model object for tracking progress of a Pipeline execution.

    Parameters
    ----------
    nterms : int
        Number of terms in the execution plan of the Pipeline being run.
    start_date : pd.Timestamp
        Start date of the range over which ``plan`` will be computed.
    end_date : pd.Timestamp
        End date of the range over which ``plan`` will be computed.

    Methods
    -------
    start_chunk(start_date, end_date)
    finish_chunk(start_date, end_date)
    load_precomputed_terms(terms)
    start_load_terms(terms)
    finish_load_terms(terms)
    start_compute_term(term)
    finish_compute_term(term)
    finish(success)

    Attributes
    ----------
    state : {'init', 'loading', 'computing', 'error', 'success'}
        Current state of the execution.
    percent_complete : float
        Percent of execution that has been completed, on a scale from 0 to 100.
    execution_time : float
        Number of seconds that the execution required. Only available if state
        is 'error' or 'success'.
    execution_bounds : (pd.Timestamp, pd.Timestamp)
        Pair of (start_date, end_date) for the entire execution.
    current_chunk_bounds : (pd.Timestamp, pd.Timestamp)
        Pair of (start_date, end_date) for the currently executing chunk.
    current_work : [ziplime.pipeline.Term]
        List of terms currently being loaded or computed.
    """

    def __init__(self, start_date, end_date):
        self._start_date = start_date
        self._end_date = end_date

        # +1 to be inclusive of end_date.
        self._total_days = (end_date - start_date).days + 1
        self._progress = 0.0
        self._days_completed = 0

        self._state = "init"

        # Number of days in current chunk.
        self._current_chunk_size = None

        # (start_date, end_date) of current chunk.
        self._current_chunk_bounds = None

        # How much should we increment progress by after completing a term?
        self._completed_term_increment = None

        # How much should we increment progress by after completing a chunk?
        # This is zero unless we compute a pipeline with no terms, in which
        # case it will be the full chunk percentage.
        self._completed_chunk_increment = None

        # Terms currently being computed.
        self._current_work = None

        # Tracking state for total elapsed time.
        self._start_time = time.time()
        self._end_time = None

    # These properties form the interface for Publishers.
    @property
    def state(self):
        return self._state

    @property
    def percent_complete(self):
        return round(self._progress * 100.0, 3)

    @property
    def execution_time(self):
        if self._end_time is None:
            raise ValueError("Can't get execution_time until execution is complete.")
        return self._end_time - self._start_time

    @property
    def execution_bounds(self):
        return (self._start_date, self._end_date)

    @property
    def current_chunk_bounds(self):
        return self._current_chunk_bounds

    @property
    def current_work(self):
        return self._current_work

    # These methods form the interface for ProgressHooks.
    def start_chunk(self, terms, start_date, end_date):
        days_since_start = (end_date - self._start_date).days + 1
        self._current_chunk_size = days_since_start - self._days_completed
        self._current_chunk_bounds = (start_date, end_date)

        # What percent of our overall progress will happen in this chunk?
        chunk_percent = float(self._current_chunk_size) / self._total_days

        # How much of that is associated with each completed term?
        nterms = len(terms)
        if nterms:
            self._completed_term_increment = chunk_percent / len(terms)
            self._completed_chunk_increment = 0.0
        else:
            # Special case. If we don't have any terms, increment the entire
            # chunk's worth of progress when we finish the chunk.
            self._completed_term_increment = 0.0
            self._completed_chunk_increment = chunk_percent

    def finish_chunk(self, terms, start_date, end_date):
        self._days_completed += self._current_chunk_size
        self._progress += self._completed_chunk_increment

    def start_load_terms(self, terms):
        self._state = "loading"
        self._current_work = terms

    def finish_load_terms(self, terms):
        self._finish_terms(nterms=len(terms))

    def start_compute_term(self, term):
        self._state = "computing"
        self._current_work = [term]

    def finish_compute_term(self, term):
        self._finish_terms(nterms=1)

    def finish(self, success):
        self._end_time = time.time()
        if success:
            self._state = "success"
        else:
            self._state = "error"

    def _finish_terms(self, nterms):
        self._progress += nterms * self._completed_term_increment


try:
    import ipywidgets

    HAVE_WIDGETS = True

    # This VBox subclass exists to work around a strange display issue but
    # where the repr of the progress bar sometimes gets re-displayed upon
    # re-opening the notebook, even after the bar has closed. The repr of VBox
    # is somewhat noisy, so we replace it here with a version that just returns
    # an empty string.
    class ProgressBarContainer(ipywidgets.VBox):
        def __repr__(self):
            return ""

except ImportError:
    HAVE_WIDGETS = False

try:
    from IPython.display import display, HTML as IPython_HTML

    HAVE_IPYTHON = True
except ImportError:
    HAVE_IPYTHON = False


# XXX: This class is currently untested, because we don't require ipywidgets as
#      a test dependency. Be careful if you make changes to this.
class IPythonWidgetProgressPublisher:
    """A progress publisher that publishes to an IPython/Jupyter widget."""

    def __init__(self):
        missing = []
        if not HAVE_WIDGETS:
            missing.append("ipywidgets")
        elif not HAVE_IPYTHON:
            missing.append("IPython")

        if missing:
            raise ValueError(
                "IPythonWidgetProgressPublisher needs ipywidgets and IPython:"
                "\nMissing:\n{}".format(bulleted_list(missing))
            )

        # Heading for progress display.
        self._heading = ipywidgets.HTML()

        # Percent Complete Indicator to the left of the bar.
        indicator_width = "120px"
        self._percent_indicator = ipywidgets.HTML(
            layout={"width": indicator_width},
        )

        # The progress bar itself.
        self._bar = ipywidgets.FloatProgress(
            value=0.0,
            min=0.0,
            max=100.0,
            bar_style="info",
            # Leave enough space for the percent indicator.
            layout={"width": "calc(100% - {})".format(indicator_width)},
        )
        bar_and_percent = ipywidgets.HBox([self._percent_indicator, self._bar])

        # Collapsable details tab underneath the progress bar.
        self._details_body = ipywidgets.HTML()
        self._details_tab = ipywidgets.Accordion(
            children=[self._details_body],
            selected_index=None,  # Start in collapsed state.
            layout={
                # Override default border settings to make details tab less
                # heavy.
                "border": "1px",
            },
        )
        # There's no public interface for setting title in the constructor :/.
        self._details_tab.set_title(0, "Details")

        # Container for the combined widget.
        self._layout = ProgressBarContainer(
            [
                self._heading,
                bar_and_percent,
                self._details_tab,
            ],
            # Overall layout consumes 75% of the page.
            layout={"width": "75%"},
        )

        self._displayed = False

    def publish(self, model):
        if model.state == "init":
            self._heading.value = "<b>Analyzing Pipeline...</b>"
            self._set_progress(0.0)
            self._ensure_displayed()

        elif model.state in ("loading", "computing"):

            term_list = self._render_term_list(model.current_work)
            if model.state == "loading":
                details_heading = "<b>Loading Inputs:</b>"
            else:
                details_heading = "<b>Computing Expression:</b>"
            self._details_body.value = details_heading + term_list

            chunk_start, chunk_end = model.current_chunk_bounds
            self._heading.value = (
                "<b>Running Pipeline</b>: Chunk Start={}, Chunk End={}".format(
                    chunk_start.date(), chunk_end.date()
                )
            )

            self._set_progress(model.percent_complete)

            self._ensure_displayed()

        elif model.state == "success":
            # Replace widget layout with html that can be persisted.
            self._stop_displaying()
            display(
                IPython_HTML(
                    "<b>Pipeline Execution Time:</b> {}".format(
                        self._format_execution_time(model.execution_time)
                    )
                ),
            )

        elif model.state == "error":
            self._bar.bar_style = "danger"
            self._stop_displaying()
        else:
            self._layout.close()
            raise ValueError("Unknown display state: {!r}".format(model.state))

    def _ensure_displayed(self):
        if not self._displayed:
            display(self._layout)
            self._displayed = True

    def _stop_displaying(self):
        self._layout.close()

    @staticmethod
    def _render_term_list(terms):
        list_elements = "".join(
            ["<li><pre>{}</pre></li>".format(repr_htmlsafe(t)) for t in terms]
        )
        return "<ul>{}</ul>".format(list_elements)

    def _set_progress(self, percent_complete):
        self._bar.value = percent_complete
        self._percent_indicator.value = "<b>{:.2f}% Complete</b>".format(
            percent_complete
        )

    @staticmethod
    def _format_execution_time(total_seconds):
        """Helper method for displaying total execution time of a Pipeline.

        Parameters
        ----------
        total_seconds : float
            Number of seconds elapsed.

        Returns
        -------
        formatted : str
            User-facing text representation of elapsed time.
        """

        def maybe_s(n):
            if n == 1:
                return ""
            return "s"

        minutes, seconds = divmod(total_seconds, 60)
        minutes = int(minutes)
        if minutes >= 60:
            hours, minutes = divmod(minutes, 60)
            t = "{hours} Hour{hs}, {minutes} Minute{ms}, {seconds:.2f} Seconds"
            return t.format(
                hours=hours,
                hs=maybe_s(hours),
                minutes=minutes,
                ms=maybe_s(minutes),
                seconds=seconds,
            )
        elif minutes >= 1:
            t = "{minutes} Minute{ms}, {seconds:.2f} Seconds"
            return t.format(
                minutes=minutes,
                ms=maybe_s(minutes),
                seconds=seconds,
            )
        else:
            return "{seconds:.2f} Seconds".format(seconds=seconds)


class TestingProgressPublisher:
    """A progress publisher that records a trace of model states for testing."""

    TraceState = namedtuple(
        "TraceState",
        [
            "state",
            "percent_complete",
            "execution_bounds",
            "current_chunk_bounds",
            "current_work",
        ],
    )

    def __init__(self):
        self.trace = []

    def publish(self, model):
        self.trace.append(
            self.TraceState(
                state=model.state,
                percent_complete=model.percent_complete,
                execution_bounds=model.execution_bounds,
                current_chunk_bounds=model.current_chunk_bounds,
                current_work=model.current_work,
            ),
        )


def repr_htmlsafe(t):
    """Repr a value and html-escape the result.

    If an error is thrown by the repr, show a placeholder.
    """
    try:
        r = repr(t)
    except Exception:
        r = "(Error Displaying {})".format(type(t).__name__)

    return escape_html(str(r), quote=True)



from ziplime.utils.compat import contextmanager as _contextmanager


# Keep track of which methods of PipelineHooks are contextmanagers. Used by
# DelegatingHooks to properly delegate to sub-hooks.
PIPELINE_HOOKS_CONTEXT_MANAGERS = set()


def contextmanager(f):
    """
    Wrapper for contextlib.contextmanager that tracks which methods of
    PipelineHooks are contextmanagers in CONTEXT_MANAGER_METHODS.
    """
    PIPELINE_HOOKS_CONTEXT_MANAGERS.add(f.__name__)
    return _contextmanager(f)


class PipelineHooks:
    """
    Interface for instrumenting SimplePipelineEngine executions.

    Methods with names like 'on_event()' should be normal methods. They will be
    called by the engine after the corresponding event.

    Methods with names like 'doing_thing()' should be context managers. They
    will be entered by the engine around the corresponding event.

    Methods
    -------
    running_pipeline(self, pipeline, start_date, end_date, chunked)
    computing_chunk(self, terms, start_date, end_date)
    loading_terms(self, terms)
    computing_term(self, term):
    """

    @contextmanager
    def running_pipeline(self, pipeline, start_date, end_date):
        """
        Contextmanager entered during execution of run_pipeline or
        run_chunked_pipeline.

        Parameters
        ----------
        pipeline : ziplime.pipeline.Pipeline
            The pipeline being executed.
        start_date : pd.Timestamp
            First date of the execution.
        end_date : pd.Timestamp
            Last date of the execution.
        """

    @contextmanager
    def computing_chunk(self, terms, start_date, end_date):
        """
        Contextmanager entered during execution of compute_chunk.

        Parameters
        ----------
        terms : list[ziplime.pipeline.Term]
            List of terms, in execution order, that will be computed. This
            value may change between chunks if ``populate_initial_workspace``
            prepopulates different terms at different times.
        start_date : pd.Timestamp
            First date of the chunk.
        end_date : pd.Timestamp
            Last date of the chunk.
        """

    @contextmanager
    def loading_terms(self, terms):
        """Contextmanager entered when loading a batch of LoadableTerms.

        Parameters
        ----------
        terms : list[ziplime.pipeline.LoadableTerm]
            Terms being loaded.
        """

    @contextmanager
    def computing_term(self, term):
        """Contextmanager entered when computing a ComputableTerm.

        Parameters
        ----------
        terms : ziplime.pipeline.ComputableTerm
            Terms being computed.
        """



class DummyMapping:
    """
    Dummy object used to provide a mapping interface for singular values.
    """

    def __init__(self, value):
        self._value = value

    def __getitem__(self, key):
        return self._value



import datetime
from abc import abstractmethod

from sqlalchemy.orm import Mapped, relationship, declared_attr

from ziplime.assets.entities.asset import Asset
from ziplime.assets.models.symbols_universe_asset import SymbolsUniverseAssetModel
from ziplime.core.db.annotated_types import AssetRouterFKPK, IntegerPK
from ziplime.core.db.base_model import BaseModel


class SymbolsUniverseModel(BaseModel):
    __tablename__ = "symbols_universe"
    id: Mapped[IntegerPK]
    assets: Mapped[list[SymbolsUniverseAssetModel]] = relationship("SymbolsUniverseAssetModel")
    name: Mapped[str]
    symbol: Mapped[str]
    universe_type: Mapped[str]

    def get_assets(self, ) -> list[Asset]:
        return []



"""
Functions for ranking and sorting.
"""
import numpy as np
from scipy.stats import rankdata
from ziplime.utils.numpy_utils import is_missing


def rankdata_1d_descending( data: np.ndarray, method: str):
    """
    1D descending version of scipy.stats.rankdata.
    """
    return rankdata(-(data.view(np.float64)), method=method)


def masked_rankdata_2d(data: np.ndarray,
                       mask: np.ndarray,
                       missing_value,
                       method: str,
                       ascending: bool):
    """
    Compute masked rankdata on data on float64, int64, or datetime64 data.
    """
    dtype_name = data.dtype.name
    if dtype_name not in ('float64', 'int64', 'datetime64[ns]'):
        raise TypeError(
            "Can't compute rankdata on array of dtype %r." % dtype_name
        )

    missing_locations = (~mask | is_missing(data, missing_value))

    # Interpret the bytes of integral data as floats for sorting.
    data = data.copy().view(np.float64)
    data[missing_locations] = np.nan
    if not ascending:
        data = -data

    # OPTIMIZATION: Fast path the default case with our own specialized
    # Cython implementation.
    if method == 'ordinal':
        result = rankdata_2d_ordinal(data)
    else:
        # FUTURE OPTIMIZATION:
        # Write a less general "apply to rows" method that doesn't do all
        # the extra work that apply_along_axis does.
        result = np.apply_along_axis(rankdata, 1, data, method=method)

        # On SciPy >= 0.17, rankdata returns integers for any method except
        # average.
        if result.dtype.name != 'float64':
            result = result.astype('float64')

    # rankdata will sort missing values into last place, but we want our nans
    # to propagate, so explicitly re-apply.
    result[missing_locations] = np.nan
    return result


def rankdata_2d_ordinal( array: np.ndarray):
    """
    Equivalent to:
    numpy.apply_over_axis(scipy.stats.rankdata, 1, array, method='ordinal')
    """
    nrows, ncols = array.shape
    sort_idxs = None
    out = None

    # scipy.stats.rankdata explicitly uses MERGESORT instead of QUICKSORT for
    # the ordinal branch.  c.f. commit ab21d2fee2d27daca0b2c161bbb7dba7e73e70ba
    # sort_idxs = np.PyArray_ArgSort(array, 1, np.NPY_MERGESORT)
    sort_idxs = np.argsort(array, axis=1, kind='mergesort')
    # Roughly, "out = np.empty_like(array)"
    # out = np.PyArray_EMPTY(2, np.PyArray_DIMS(array), np.NPY_DOUBLE, False)
    out = np.empty(array.shape, dtype=np.float64)
    for i in range(nrows):
        for j in range(ncols):
            out[i, sort_idxs[i, j]] = j + 1.0

    return out


def grouped_masked_is_maximal(data: np.ndarray, # 2 dim
                                groupby,
                                mask):
    """Build a mask of the top value for each row in ``data``, grouped by
    ``groupby`` and masked by ``mask``.
    Parameters
    ----------
    data : np.array[np.int64_t]
        Data on which we should find maximal values for each row.
    groupby : np.array[np.int64_t]
        Grouping labels for rows of ``data``. We choose one entry in each
        row for each unique grouping key in that row.
    mask : np.array[np.uint8_t]
        Boolean mask of locations to consider as possible maximal values.
        Locations with a 0 in ``mask`` are ignored.
    Returns
    -------
    maximal_locations : np.array[bool]
        Mask containing True for the maximal non-masked value in each row/group.
    """
    # Cython thinks ``.shape`` is an intp_t pointer on ndarrays, so we need to
    # cast to object to get the proper shape attribute.
    if not (data.shape == groupby.shape == data.shape):
        raise AssertionError(
            "Misaligned shapes in grouped_masked_is_maximal:"
            "data={}, groupby={}, mask={}".format(
                data.shape, groupby.shape, mask.shape,
            )
        )

    group = None
    value = None
    out = np.zeros_like(mask)
    best_per_group = {}
    nrows, ncols = data.shape
    for i in range(nrows):
        best_per_group.clear()
        for j in range(ncols):

            # NOTE: Callers are responsible for masking out values that should
            # be treated as null here.
            if not mask[i, j]:
                continue

            value = data[i, j]
            group = groupby[i, j]

            if group not in best_per_group:
                best_per_group[group] = j
                continue

            if value > data[i, best_per_group[group]]:
                best_per_group[group] = j

        for j in best_per_group.values():
            out[i, j] = 1

    return out.view(bool)



from enum import Enum


class OrderType(Enum):
    """
    Enum representing all possible order types in trading systems.
    """
    # Basic order types
    MARKET = "MARKET"  # Execute immediately at current market price
    LIMIT = "LIMIT"  # Buy/sell at a specified price or better
    LIMIT_MAKER = "LIMIT_MAKER"  # Like limit but ensures the order is a maker

    # Additional order types
    STOP = "STOP"  # Market order that activates at a specified trigger price
    STOP_LIMIT = "STOP_LIMIT"  # Limit order that activates at a specified trigger price
    TRAILING_STOP = "TRAILING_STOP"  # Stop order where trigger price adjusts with market movement
    FILL_OR_KILL = "FILL_OR_KILL"  # Must be filled entirely immediately or canceled entirely
    IMMEDIATE_OR_CANCEL = "IMMEDIATE_OR_CANCEL"  # Must be filled immediately (partial fill allowed) or canceled
    GOOD_TILL_CANCELLED = "GOOD_TILL_CANCELLED"  # Order stays active until manually canceled
    GOOD_TILL_DATE = "GOOD_TILL_DATE"  # Order stays active until a specified date/time
    ICEBERG = "ICEBERG"  # Large order that shows only a small portion to the market
    ONE_CANCELS_OTHER = "ONE_CANCELS_OTHER"  # Two linked orders where execution of one cancels the other
    POST_ONLY = "POST_ONLY"  # Order is posted to the book and not matched immediately
    BRACKET = "BRACKET"  # Set of orders including entry, take profit, and stop loss
    CONDITIONAL = "CONDITIONAL"  # Order that executes based on specified market conditions
    PEGGED = "PEGGED"  # Order that adjusts price relative to a reference price
    REDUCE_ONLY = "REDUCE_ONLY"  # Order that can only reduce, not increase, a position
    TAKE_PROFIT = "TAKE_PROFIT"  # Order to close a position at a profit target
    TAKE_PROFIT_LIMIT = "TAKE_PROFIT_LIMIT"  # Limit order version of take profit

    def is_limit_type(self) -> bool:
        """
        Determines if the order type is a limit-based order.
        """
        return self in (OrderType.LIMIT, OrderType.LIMIT_MAKER, OrderType.STOP_LIMIT,
                        OrderType.TAKE_PROFIT_LIMIT, OrderType.POST_ONLY)

    def is_stop_type(self) -> bool:
        """
        Determines if the order type is a stop-based order.
        """
        return self in (OrderType.STOP, OrderType.STOP_LIMIT, OrderType.TRAILING_STOP)

    def requires_price(self) -> bool:
        """
        Determines if the order type requires a price parameter.
        """
        return self.is_limit_type() or self.is_stop_type()



import datetime

import structlog
from exchange_calendars import ExchangeCalendar

import polars as pl

from ziplime.assets.services.asset_service import AssetService
from ziplime.constants.data_type import DataType
from ziplime.constants.period import Period
from ziplime.data.services.data_source import DataSource
from ziplime.utils.data_utils import _process_data


class CSVDataSource(DataSource):
    def __init__(self,
                 name: str,
                 csv_file_name: str,
                 column_mapping: dict[str, str],
                 frequency: datetime.timedelta | Period,
                 date_column_name: str,
                 date_format: str,
                 data_frequency_use_window_end: bool,
                 symbols: list[str],
                 asset_service: AssetService,
                 trading_calendar: ExchangeCalendar,
                 data_type: DataType,
                 ):
        self._csv_file_name = csv_file_name
        self._column_mapping = column_mapping
        self._logger = structlog.get_logger(__name__)
        self._date_format = date_format
        self._date_column_name = date_column_name
        self._asset_service = asset_service
        self._trading_calendar = trading_calendar
        self._data_frequency_use_window_end = data_frequency_use_window_end
        self._symbols = symbols
        self.data = None
        self.start_date = None
        self.end_date = None
        self.data_type = data_type
        super().__init__(name=name, start_date=self.start_date, end_date=self.end_date, frequency=frequency,
                         data_type=data_type, original_frequency=frequency)

    async def load_data_in_memory(self) -> pl.DataFrame:
        df = pl.read_csv(self._csv_file_name)  # s, schema_overrides={self._date_column_name: pl.Datetime})

        # Parse "date" column with your format
        df = df.with_columns(
            pl.col(self._date_column_name).str.strptime(pl.Datetime("us"), format=self._date_format).alias(
                self._date_column_name)
        )

        df = df.rename(self._column_mapping)
        df = df.with_columns(
            pl.col("close").alias("price"),
            date=pl.col("date").dt.replace_time_zone(str(self._trading_calendar.tz))
        )

        df = await _process_data(data=df,
                           date_start=df["date"].min(),
                           date_end=df["date"].max(),
                           data_frequency_use_window_end=self._data_frequency_use_window_end,
                           frequency=self.frequency,
                           trading_calendar=self._trading_calendar,
                           asset_service=self._asset_service,
                           name=self.name,
                           symbols=self._symbols,

                           )
        self.data = df
        self.start_date = self.data["date"].min()
        self.end_date = self.data["date"].max()



import os

from ziplime.data.services.lime_trader_sdk_data_source import LimeTraderSdkDataSource
from ziplime.exchanges.lime_trader_sdk.lime_trader_sdk_exchange import LimeTraderSdkExchange
from ziplime.data.providers.fundamental_data.limex_hub_fundamental_data_provider import LimexHubFundamentalDataProvider


def get_data_source(code: str):
    if code == "lime-trader-sdk":
        lime_trader_sdk_credentials = os.environ.get("LIME_SDK_CREDENTIALS_FILE", None)
        if lime_trader_sdk_credentials is None:
            raise ValueError("Missing LIME_SDK_CREDENTIALS_FILE environment variable.")
        return LimeTraderSdkDataSource(lime_sdk_credentials_file=lime_trader_sdk_credentials)
    raise Exception("Unsupported live market data provider!")

def get_fundamental_data_provider(code: str):
    if code == "limex-hub":
        limex_hub_key = os.environ.get("LIMEX_API_KEY", None)
        maximum_threads = os.environ.get("LIMEX_HUB_MAXIMUM_THREADS", None)
        if limex_hub_key is None:
            raise ValueError("Missing LIMEX_API_KEY environment variable.")
        return LimexHubFundamentalDataProvider(limex_api_key=limex_hub_key, maximum_threads=maximum_threads)
    raise Exception("Unsupported fundamental data provider!")

def get_exchange(code: str):
    if code == "lime-trader-sdk":
        lime_trader_sdk_credentials = os.environ.get("LIME_SDK_CREDENTIALS_FILE", None)
        if lime_trader_sdk_credentials is None:
            raise ValueError("Missing LIME_SDK_CREDENTIALS_FILE environment variable.")
        return LimeTraderSdkExchange(lime_sdk_credentials_file=lime_trader_sdk_credentials)
    raise Exception("Unsupported live market data provider!")



import datetime

import aiocache
import polars as pl
from aiocache import Cache

from ziplime.assets.domain.asset_type import AssetType
from ziplime.assets.entities.asset import Asset
from ziplime.assets.entities.commodity import Commodity
from ziplime.assets.entities.currency import Currency
from ziplime.assets.entities.equity import Equity
from ziplime.assets.entities.futures_contract import FuturesContract
from ziplime.assets.entities.symbol_universe import SymbolsUniverse
from ziplime.assets.models.dividend import Dividend
from ziplime.assets.repositories.adjustments_repository import AdjustmentRepository
from ziplime.assets.repositories.asset_repository import AssetRepository
from ziplime.exchanges.exchange import Exchange
from ziplime.trading.entities.trading_pair import TradingPair


class AssetService:

    def __init__(self, asset_repository: AssetRepository, adjustments_repository: AdjustmentRepository):
        self._asset_repository = asset_repository
        self._adjustments_repository = adjustments_repository

    async def save_equities(self, equities: list[Equity]) -> None:
        await self._asset_repository.save_equities(equities=equities)
    async def save_symbol_universe(self, symbol_universe: SymbolsUniverse) -> None:
        await self._asset_repository.save_symbol_universe(symbol_universe=symbol_universe)

    async def save_commodities(self, commodities: list[Commodity]) -> None:
        await self._asset_repository.save_commodities(commodities=commodities)

    async def save_currencies(self, currencies: list[Currency]) -> None:
        await self._asset_repository.save_currencies(currencies=currencies)

    async def save_exchanges(self, exchanges: list[Exchange]) -> None:
        return await self._asset_repository.save_exchanges(exchanges=exchanges)

    async def save_trading_pairs(self, trading_pairs: list[TradingPair]) -> None: ...

    async def get_asset_by_sid(self, sid: int) -> Asset | None:
        return await self._asset_repository.get_asset_by_sid(sid=sid)

    async def get_assets_by_sids(self, sids: list[int]) -> list[Asset]:
        return await self._asset_repository.get_assets_by_sids(sids=sids)

    async def get_equity_by_symbol(self, symbol: str, exchange_name: str) -> Equity | None:
        return await self._asset_repository.get_equity_by_symbol(symbol=symbol,
                                                                 exchange_name=exchange_name)

    async def get_equities_by_symbols(self, symbols: list[str]) -> list[Equity]:
        return await self._asset_repository.get_equities_by_symbols(symbols=symbols)

    async def get_equities_by_symbols_and_exchange(self, symbols: list[str], exchange_name: str) -> list[Equity]:
        return await self._asset_repository.get_equities_by_symbols_and_exchange(symbols=symbols,
                                                                                 exchange_name=exchange_name)

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_asset_by_symbol(self, symbol: str, asset_type: AssetType, exchange_name: str) -> Asset | None:
        return await self._asset_repository.get_asset_by_symbol(symbol=symbol,
                                                                asset_type=asset_type,
                                                                exchange_name=exchange_name)

    async def get_futures_contract_by_symbol(self, symbol: str, exchange_name: str) -> FuturesContract | None:
        return await self._asset_repository.get_futures_contract_by_symbol(symbol=symbol,
                                                                           exchange_name=exchange_name)

    @aiocache.cached(cache=Cache.MEMORY)
    async def get_currency_by_symbol(self, symbol: str, exchange_name: str) -> Currency | None:
        return await self._asset_repository.get_currency_by_symbol(symbol=symbol,
                                                                   exchange_name=exchange_name)

    async def get_commodity_by_symbol(self, symbol: str, exchange_name: str) -> Commodity | None:
        return await self._asset_repository.get_commodity_by_symbol(symbol=symbol,
                                                                    exchange_name=exchange_name)

    def get_stock_dividends(self, sid: int, trading_days: pl.Series) -> list[Dividend]:
        return self._adjustments_repository.get_stock_dividends(sid=sid,
                                                                trading_days=trading_days)

    def get_splits(self, assets: frozenset[Asset], dt: datetime.date):
        return self._adjustments_repository.get_splits(assets=assets, dt=dt)

    async def get_symbols_universe(self, symbol: str) -> SymbolsUniverse | None:
        return await self._asset_repository.get_symbols_universe(symbol=symbol)



#
# Copyright 2012 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

TRADING_DAYS_IN_YEAR = 250
TRADING_HOURS_IN_DAY = 6.5
MINUTES_IN_HOUR = 60


# NOTE: It may be worth revisiting how the keys for this dictionary are
# specified, for instance making them ContinuousFuture objects instead of
# static strings.
FUTURE_EXCHANGE_FEES_BY_SYMBOL = {
    "AD": 1.60,  # AUD
    "AI": 0.96,  # Bloomberg Commodity Index
    "BD": 1.50,  # Big Dow
    "BO": 1.95,  # Soybean Oil
    "BP": 1.60,  # GBP
    "CD": 1.60,  # CAD
    "CL": 1.50,  # Crude Oil
    "CM": 1.03,  # Corn e-mini
    "CN": 1.95,  # Corn
    "DJ": 1.50,  # Dow Jones
    "EC": 1.60,  # Euro FX
    "ED": 1.25,  # Eurodollar
    "EE": 1.50,  # Euro FX e-mini
    "EI": 1.50,  # MSCI Emerging Markets mini
    "EL": 1.50,  # Eurodollar NYSE LIFFE
    "ER": 0.65,  # Russell2000 e-mini
    "ES": 1.18,  # SP500 e-mini
    "ET": 1.50,  # Ethanol
    "EU": 1.50,  # Eurodollar e-micro
    "FC": 2.03,  # Feeder Cattle
    "FF": 0.96,  # 3-Day Federal Funds
    "FI": 0.56,  # Deliverable Interest Rate Swap 5y
    "FS": 1.50,  # Interest Rate Swap 5y
    "FV": 0.65,  # US 5y
    "GC": 1.50,  # Gold
    "HG": 1.50,  # Copper
    "HO": 1.50,  # Heating Oil
    "HU": 1.50,  # Unleaded Gasoline
    "JE": 0.16,  # JPY e-mini
    "JY": 1.60,  # JPY
    "LB": 2.03,  # Lumber
    "LC": 2.03,  # Live Cattle
    "LH": 2.03,  # Lean Hogs
    "MB": 1.50,  # Municipal Bonds
    "MD": 1.50,  # SP400 Midcap
    "ME": 1.60,  # MXN
    "MG": 1.50,  # MSCI EAFE mini
    "MI": 1.18,  # SP400 Midcap e-mini
    "MS": 1.03,  # Soybean e-mini
    "MW": 1.03,  # Wheat e-mini
    "ND": 1.50,  # Nasdaq100
    "NG": 1.50,  # Natural Gas
    "NK": 2.15,  # Nikkei225
    "NQ": 1.18,  # Nasdaq100 e-mini
    "NZ": 1.60,  # NZD
    "OA": 1.95,  # Oats
    "PA": 1.50,  # Palladium
    "PB": 1.50,  # Pork Bellies
    "PL": 1.50,  # Platinum
    "QG": 0.50,  # Natural Gas e-mini
    "QM": 1.20,  # Crude Oil e-mini
    "RM": 1.50,  # Russell1000 e-mini
    "RR": 1.95,  # Rough Rice
    "SB": 2.10,  # Sugar
    "SF": 1.60,  # CHF
    "SM": 1.95,  # Soybean Meal
    "SP": 2.40,  # SP500
    "SV": 1.50,  # Silver
    "SY": 1.95,  # Soybean
    "TB": 1.50,  # Treasury Bills
    "TN": 0.56,  # Deliverable Interest Rate Swap 10y
    "TS": 1.50,  # Interest Rate Swap 10y
    "TU": 1.50,  # US 2y
    "TY": 0.75,  # US 10y
    "UB": 0.85,  # Ultra Tbond
    "US": 0.80,  # US 30y
    "VX": 1.50,  # VIX
    "WC": 1.95,  # Wheat
    "XB": 1.50,  # RBOB Gasoline
    "XG": 0.75,  # Gold e-mini
    "YM": 1.50,  # Dow Jones e-mini
    "YS": 0.75,  # Silver e-mini
}

# See `ziplime.finance.slippage.VolatilityVolumeShare` for more information on
# how these constants are used.
DEFAULT_ETA = 0.049018143225019836
ROOT_SYMBOL_TO_ETA = {
    "AD": DEFAULT_ETA,  # AUD
    "AI": DEFAULT_ETA,  # Bloomberg Commodity Index
    "BD": 0.050346811117733474,  # Big Dow
    "BO": 0.054930995070046298,  # Soybean Oil
    "BP": 0.047841544238716338,  # GBP
    "CD": 0.051124420640250717,  # CAD
    "CL": 0.04852544628414196,  # Crude Oil
    "CM": 0.052683478163348625,  # Corn e-mini
    "CN": 0.053499718390037809,  # Corn
    "DJ": 0.02313009072076987,  # Dow Jones
    "EC": 0.04885131067661861,  # Euro FX
    "ED": 0.094184297090245755,  # Eurodollar
    "EE": 0.048713151357687556,  # Euro FX e-mini
    "EI": 0.031712708439692663,  # MSCI Emerging Markets mini
    "EL": 0.044207422018209361,  # Eurodollar NYSE LIFFE
    "ER": 0.045930567737711307,  # Russell2000 e-mini
    "ES": 0.047304418321993502,  # SP500 e-mini
    "ET": DEFAULT_ETA,  # Ethanol
    "EU": 0.049750396084029064,  # Eurodollar e-micro
    "FC": 0.058728734202178494,  # Feeder Cattle
    "FF": 0.048970591527624042,  # 3-Day Federal Funds
    "FI": 0.033477176738170772,  # Deliverable Interest Rate Swap 5y
    "FS": 0.034557788010453824,  # Interest Rate Swap 5y
    "FV": 0.046544427716056963,  # US 5y
    "GC": 0.048933313546125207,  # Gold
    "HG": 0.052238417524987799,  # Copper
    "HO": 0.045061318412156062,  # Heating Oil
    "HU": 0.017154313062463938,  # Unleaded Gasoline
    "JE": 0.013948949613401812,  # JPY e-mini
    "JY": DEFAULT_ETA,  # JPY
    "LB": 0.06146586386903994,  # Lumber
    "LC": 0.055853801862858619,  # Live Cattle
    "LH": 0.057557004630219781,  # Lean Hogs
    "MB": DEFAULT_ETA,  # Municipal Bonds
    "MD": DEFAULT_ETA,  # SP400 Midcap
    "ME": 0.030383767727818548,  # MXN
    "MG": 0.029579261656151684,  # MSCI EAFE mini
    "MI": 0.041026288873007355,  # SP400 Midcap e-mini
    "MS": DEFAULT_ETA,  # Soybean e-mini
    "MW": 0.052579919663880245,  # Wheat e-mini
    "ND": DEFAULT_ETA,  # Nasdaq100
    "NG": 0.047897809233755716,  # Natural Gas
    "NK": 0.044555435054791433,  # Nikkei225
    "NQ": 0.044772425085977945,  # Nasdaq100 e-mini
    "NZ": 0.049170418073872041,  # NZD
    "OA": 0.056973267232775522,  # Oats
    "PA": DEFAULT_ETA,  # Palladium
    "PB": DEFAULT_ETA,  # Pork Bellies
    "PL": 0.054579379665647493,  # Platinum
    "QG": DEFAULT_ETA,  # Natural Gas e-mini
    "QM": DEFAULT_ETA,  # Crude Oil e-mini
    "RM": 0.037425041244579654,  # Russell1000 e-mini
    "RR": DEFAULT_ETA,  # Rough Rice
    "SB": 0.057388160345668134,  # Sugar
    "SF": 0.047784825569615726,  # CHF
    "SM": 0.048552860559844223,  # Soybean Meal
    "SP": DEFAULT_ETA,  # SP500
    "SV": 0.052691435039931109,  # Silver
    "SY": 0.052041703657281613,  # Soybean
    "TB": DEFAULT_ETA,  # Treasury Bills
    "TN": 0.033363465365262503,  # Deliverable Interest Rate Swap 10y
    "TS": 0.032908878455069152,  # Interest Rate Swap 10y
    "TU": 0.063867646063840794,  # US 2y
    "TY": 0.050586988554700826,  # US 10y
    "UB": DEFAULT_ETA,  # Ultra Tbond
    "US": 0.047984179873590722,  # US 30y
    "VX": DEFAULT_ETA,  # VIX
    "WC": 0.052636542119329242,  # Wheat
    "XB": 0.044444916388854484,  # RBOB Gasoline
    "XG": DEFAULT_ETA,  # Gold e-mini
    "YM": DEFAULT_ETA,  # Dow Jones e-mini
    "YS": DEFAULT_ETA,  # Silver e-mini
}



import datetime
import os
from pathlib import Path
from typing import Any

import aiofiles.os
import orjson
import structlog

from ziplime.data.domain.data_bundle import DataBundle
from ziplime.data.services.bundle_registry import BundleRegistry
from ziplime.data.services.bundle_storage import BundleStorage


class FileSystemBundleRegistry(BundleRegistry):

    def __init__(self, base_data_path: str):
        super().__init__()
        self._base_data_path = base_data_path
        self._logger = structlog.get_logger(__name__)
        os.makedirs(self._base_data_path, exist_ok=True)

    async def get_bundle_metadata(self, data_bundle: DataBundle, bundle_storage: BundleStorage) -> dict[str, Any]:
        frequency_seconds = None
        frequency_text = None
        if type(data_bundle.frequency) is datetime.timedelta:
            frequency_seconds = data_bundle.frequency.total_seconds()
        else:
            frequency_text = data_bundle.frequency
        return {
            "name": data_bundle.name,
            "version": data_bundle.version,

            "bundle_storage_class": f"{bundle_storage.__class__.__module__}.{bundle_storage.__class__.__name__}",
            "bundle_storage_data": await bundle_storage.to_json(data_bundle=data_bundle),

            # "asset_repository_class": f"{data_bundle.asset_repository.__class__.__module__}.{data_bundle.asset_repository.__class__.__name__}",
            # "asset_repository_data": data_bundle.asset_repository.to_json(),
            #
            # "adjustment_repository_class": f"{data_bundle.adjustment_repository.__class__.__module__}.{data_bundle.adjustment_repository.__class__.__name__}",
            # "adjustment_repository_data": data_bundle.adjustment_repository.to_json(),

            "start_date": data_bundle.start_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "end_date": data_bundle.end_date.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "trading_calendar_name": data_bundle.trading_calendar.name,
            "frequency_seconds": frequency_seconds,
            "frequency_text": frequency_text,
            "timestamp": data_bundle.timestamp.strftime("%Y-%m-%dT%H:%M:%SZ"),
            "data_type": data_bundle.data_type.value,
        }

    async def load_bundle_metadata(self, bundle_name: str, bundle_version: str | None) -> dict[str, Any] | None:
        if bundle_version is None:
            bundles = await self.list_bundles_by_name(bundle_name=bundle_name)
            if not bundles:
                return None
            bundle_version = bundles[0]["version"]

        bundle_metadata_path = Path(self.get_bundle_registry_path(), f"{bundle_name}_{bundle_version}.json")
        async with aiofiles.open(bundle_metadata_path, mode="rb") as f:
            return orjson.loads(await f.read())

    def get_bundle_registry_path(self) -> Path:
        return Path(self._base_data_path, "bundle_registry")

    async def persist_metadata(self, data_bundle: DataBundle, metadata: dict[str, Any]):
        bundle_metadata_path = Path(self.get_bundle_registry_path(), f"{data_bundle.name}_{data_bundle.version}.json")
        await aiofiles.os.makedirs(bundle_metadata_path.parent, exist_ok=True)
        async with aiofiles.open(bundle_metadata_path, mode="wb") as f:
            await f.write(orjson.dumps(metadata, option=orjson.OPT_INDENT_2))

    async def delete_bundle(self):
        pass

    async def list_bundles(self) -> list[dict[str, Any]]:
        registry_items = []
        for file in await aiofiles.os.scandir(self.get_bundle_registry_path()):
            if file.is_file():
                if not file.name.endswith(".json"):
                    continue
                registry_items.append(file.path)
        bundles = []
        for item in registry_items:
            async with aiofiles.open(item, mode="r") as f:
                bundle_metadata = orjson.loads(await f.read())
                bundles.append(bundle_metadata)
        return bundles

    async def list_bundles_by_name(self, bundle_name: str) -> list[dict[str, Any]]:
        bundles = await self.list_bundles()
        return sorted(list(filter(lambda b: b["name"] == bundle_name, bundles)), key=lambda b: b["timestamp"],
                      reverse=True)



"""
Technical Analysis Factors
--------------------------
"""

from numpy import (
    abs,
    average,
    clip,
    diff,
    dstack,
    inf,
)
from numexpr import evaluate

from ziplime.pipeline.data import EquityPricing
from ziplime.pipeline.factors import CustomFactor
from ziplime.pipeline.mixins import SingleInputMixin
from ziplime.utils.math_utils import (
    nanargmax,
    nanargmin,
    nanmax,
    nanmean,
    nanstd,
    nanmin,
)
from ziplime.utils.numpy_utils import rolling_window

from .basic import exponential_weights
from .basic import (  # noqa reexport
    # These are re-exported here for backwards compatibility with the old
    # definition site.
    LinearWeightedMovingAverage,
    MaxDrawdown,
    SimpleMovingAverage,
    VWAP,
    WeightedAverageValue,
)


class RSI(SingleInputMixin, CustomFactor):
    """
    Relative Strength Index

    **Default Inputs**: :data:`ziplime.pipeline.data.EquityPricing.close`

    **Default Window Length**: 15
    """

    window_length = 15
    inputs = (EquityPricing.close,)
    window_safe = True

    def compute(self, today, assets, out, closes):
        diffs = diff(closes, axis=0)
        ups = nanmean(clip(diffs, 0, inf), axis=0)
        downs = abs(nanmean(clip(diffs, -inf, 0), axis=0))
        return evaluate(
            "100 - (100 / (1 + (ups / downs)))",
            local_dict={"ups": ups, "downs": downs},
            global_dict={},
            out=out,
        )


class BollingerBands(CustomFactor):
    """
    Bollinger Bands technical indicator.
    https://en.wikipedia.org/wiki/Bollinger_Bands

    **Default Inputs:** :data:`ziplime.pipeline.data.EquityPricing.close`

    Parameters
    ----------
    inputs : length-1 iterable[BoundColumn]
        The expression over which to compute bollinger bands.
    window_length : int > 0
        Length of the lookback window over which to compute the bollinger
        bands.
    k : float
        The number of standard deviations to add or subtract to create the
        upper and lower bands.
    """

    params = ("k",)
    inputs = (EquityPricing.close,)
    outputs = "lower", "middle", "upper"

    def compute(self, today, assets, out, close, k):
        difference = k * nanstd(close, axis=0)
        out.middle = middle = nanmean(close, axis=0)
        out.upper = middle + difference
        out.lower = middle - difference


class Aroon(CustomFactor):
    """
    Aroon technical indicator.
    https://www.fidelity.com/learning-center/trading-investing/technical-analysis/technical-indicator-guide/aroon-indicator

    **Defaults Inputs:** :data:`ziplime.pipeline.data.EquityPricing.low`, \
                         :data:`ziplime.pipeline.data.EquityPricing.high`

    Parameters
    ----------
    window_length : int > 0
        Length of the lookback window over which to compute the Aroon
        indicator.
    """  # noqa

    inputs = (EquityPricing.low, EquityPricing.high)
    outputs = ("down", "up")

    def compute(self, today, assets, out, lows, highs):
        wl = self.window_length
        high_date_index = nanargmax(highs, axis=0)
        low_date_index = nanargmin(lows, axis=0)
        evaluate(
            "(100 * high_date_index) / (wl - 1)",
            local_dict={
                "high_date_index": high_date_index,
                "wl": wl,
            },
            out=out.up,
        )
        evaluate(
            "(100 * low_date_index) / (wl - 1)",
            local_dict={
                "low_date_index": low_date_index,
                "wl": wl,
            },
            out=out.down,
        )


class FastStochasticOscillator(CustomFactor):
    """
    Fast Stochastic Oscillator Indicator [%K, Momentum Indicator]
    https://wiki.timetotrade.eu/Stochastic

    This stochastic is considered volatile, and varies a lot when used in
    market analysis. It is recommended to use the slow stochastic oscillator
    or a moving average of the %K [%D].

    **Default Inputs:** :data:`ziplime.pipeline.data.EquityPricing.close`, \
                        :data:`ziplime.pipeline.data.EquityPricing.low`, \
                        :data:`ziplime.pipeline.data.EquityPricing.high`

    **Default Window Length:** 14

    Returns
    -------
    out: %K oscillator
    """

    inputs = (EquityPricing.close, EquityPricing.low, EquityPricing.high)
    window_safe = True
    window_length = 14

    def compute(self, today, assets, out, closes, lows, highs):

        highest_highs = nanmax(highs, axis=0)
        lowest_lows = nanmin(lows, axis=0)
        today_closes = closes[-1]

        evaluate(
            "((tc - ll) / (hh - ll)) * 100",
            local_dict={
                "tc": today_closes,
                "ll": lowest_lows,
                "hh": highest_highs,
            },
            global_dict={},
            out=out,
        )


class IchimokuKinkoHyo(CustomFactor):
    """Compute the various metrics for the Ichimoku Kinko Hyo (Ichimoku Cloud).
    http://stockcharts.com/school/doku.php?id=chart_school:technical_indicators:ichimoku_cloud

    **Default Inputs:** :data:`ziplime.pipeline.data.EquityPricing.high`, \
                        :data:`ziplime.pipeline.data.EquityPricing.low`, \
                        :data:`ziplime.pipeline.data.EquityPricing.close`

    **Default Window Length:** 52

    Parameters
    ----------
    window_length : int > 0
        The length the the window for the senkou span b.
    tenkan_sen_length : int >= 0, <= window_length
        The length of the window for the tenkan-sen.
    kijun_sen_length : int >= 0, <= window_length
        The length of the window for the kijou-sen.
    chikou_span_length : int >= 0, <= window_length
        The lag for the chikou span.
    """  # noqa

    params = {
        "tenkan_sen_length": 9,
        "kijun_sen_length": 26,
        "chikou_span_length": 26,
    }
    inputs = (EquityPricing.high, EquityPricing.low, EquityPricing.close)
    outputs = (
        "tenkan_sen",
        "kijun_sen",
        "senkou_span_a",
        "senkou_span_b",
        "chikou_span",
    )
    window_length = 52

    def _validate(self):
        super(IchimokuKinkoHyo, self)._validate()
        for k, v in self.params.items():
            if v > self.window_length:
                raise ValueError(
                    "%s must be <= the window_length: %s > %s"
                    % (
                        k,
                        v,
                        self.window_length,
                    ),
                )

    def compute(
        self,
        today,
        assets,
        out,
        high,
        low,
        close,
        tenkan_sen_length,
        kijun_sen_length,
        chikou_span_length,
    ):

        out.tenkan_sen = tenkan_sen = (
            high[-tenkan_sen_length:].max(axis=0) + low[-tenkan_sen_length:].min(axis=0)
        ) / 2
        out.kijun_sen = kijun_sen = (
            high[-kijun_sen_length:].max(axis=0) + low[-kijun_sen_length:].min(axis=0)
        ) / 2
        out.senkou_span_a = (tenkan_sen + kijun_sen) / 2
        out.senkou_span_b = (high.max(axis=0) + low.min(axis=0)) / 2
        out.chikou_span = close[chikou_span_length]


class RateOfChangePercentage(CustomFactor):
    """
    Rate of change Percentage
    ROC measures the percentage change in price from one period to the next.
    The ROC calculation compares the current price with the price `n`
    periods ago.
    Formula for calculation: ((price - prevPrice) / prevPrice) * 100
    price - the current price
    prevPrice - the price n days ago, equals window length
    """

    def compute(self, today, assets, out, close):
        today_close = close[-1]
        prev_close = close[0]
        evaluate(
            "((tc - pc) / pc) * 100",
            local_dict={"tc": today_close, "pc": prev_close},
            global_dict={},
            out=out,
        )


class TrueRange(CustomFactor):
    """
    True Range

    A technical indicator originally developed by J. Welles Wilder, Jr.
    Indicates the true degree of daily price change in an underlying.

    **Default Inputs:** :data:`ziplime.pipeline.data.EquityPricing.high`, \
                        :data:`ziplime.pipeline.data.EquityPricing.low`, \
                        :data:`ziplime.pipeline.data.EquityPricing.close`

    **Default Window Length:** 2
    """

    inputs = (
        EquityPricing.high,
        EquityPricing.low,
        EquityPricing.close,
    )
    window_length = 2

    def compute(self, today, assets, out, highs, lows, closes):
        high_to_low = highs[1:] - lows[1:]
        high_to_prev_close = abs(highs[1:] - closes[:-1])
        low_to_prev_close = abs(lows[1:] - closes[:-1])
        out[:] = nanmax(
            dstack(
                (
                    high_to_low,
                    high_to_prev_close,
                    low_to_prev_close,
                )
            ),
            2,
        )


class MovingAverageConvergenceDivergenceSignal(CustomFactor):
    """
    Moving Average Convergence/Divergence (MACD) Signal line
    https://en.wikipedia.org/wiki/MACD

    A technical indicator originally developed by Gerald Appel in the late
    1970's. MACD shows the relationship between two moving averages and
    reveals changes in the strength, direction, momentum, and duration of a
    trend in a stock's price.

    **Default Inputs:** :data:`ziplime.pipeline.data.EquityPricing.close`

    Parameters
    ----------
    fast_period : int > 0, optional
        The window length for the "fast" EWMA. Default is 12.
    slow_period : int > 0, > fast_period, optional
        The window length for the "slow" EWMA. Default is 26.
    signal_period : int > 0, < fast_period, optional
        The window length for the signal line. Default is 9.

    Notes
    -----
    Unlike most pipeline expressions, this factor does not accept a
    ``window_length`` parameter. ``window_length`` is inferred from
    ``slow_period`` and ``signal_period``.
    """

    inputs = (EquityPricing.close,)
    # We don't use the default form of `params` here because we want to
    # dynamically calculate `window_length` from the period lengths in our
    # __new__.
    params = ("fast_period", "slow_period", "signal_period")

    def __new__(cls, fast_period=12, slow_period=26, signal_period=9, *args, **kwargs):
        if fast_period < 1:
            raise ValueError("`fast_period` must be >= 1")
        if slow_period < 1:
            raise ValueError("`slow_period` must be >= 1")
        if signal_period < 1:
            raise ValueError("`signal_period` must be >= 1")

        if slow_period <= fast_period:
            raise ValueError(
                "'slow_period' must be greater than 'fast_period', but got\n"
                "slow_period={slow}, fast_period={fast}".format(
                    slow=slow_period,
                    fast=fast_period,
                )
            )

        return super(MovingAverageConvergenceDivergenceSignal, cls).__new__(
            cls,
            fast_period=fast_period,
            slow_period=slow_period,
            signal_period=signal_period,
            window_length=slow_period + signal_period - 1,
            *args,
            **kwargs,
        )

    def _ewma(self, data, length):
        decay_rate = 1.0 - (2.0 / (1.0 + length))
        return average(data, axis=1, weights=exponential_weights(length, decay_rate))

    def compute(
        self, today, assets, out, close, fast_period, slow_period, signal_period
    ):
        slow_EWMA = self._ewma(rolling_window(close, slow_period), slow_period)
        fast_EWMA = self._ewma(
            rolling_window(close, fast_period)[-signal_period:], fast_period
        )
        macd = fast_EWMA - slow_EWMA
        out[:] = self._ewma(macd.T, signal_period)


# Convenience aliases.
MACDSignal = MovingAverageConvergenceDivergenceSignal



from enum import Enum


class TimeInForce(Enum):
    """
    Enum representing time-in-force parameters for orders.
    """
    GTC = "GTC"  # Good Till Cancelled
    IOC = "IOC"  # Immediate Or Cancel
    FOK = "FOK"  # Fill Or Kill
    GTD = "GTD"  # Good Till Date
    DAY = "DAY"  # Valid for the trading day
    AT_THE_OPENING = "OPG"  # Execute at market opening
    AT_THE_CLOSE = "CLS"  # Execute at market close



import numpy as np


def naive_grouped_rowwise_apply(data, group_labels, func, func_args=(), out=None):
    """
    Simple implementation of grouped row-wise function application.

    Parameters
    ----------
    data : ndarray[ndim=2]
        Input array over which to apply a grouped function.
    group_labels : ndarray[ndim=2, dtype=int64]
        Labels to use to bucket inputs from array.
        Should be the same shape as array.
    func : function[ndarray[ndim=1]] -> function[ndarray[ndim=1]]
        Function to apply to pieces of each row in array.
    func_args : tuple
        Additional positional arguments to provide to each row in array.
    out : ndarray, optional
        Array into which to write output.  If not supplied, a new array of the
        same shape as ``data`` is allocated and returned.

    Examples
    --------
    >>> data = np.array([[1., 2., 3.],
    ...                  [2., 3., 4.],
    ...                  [5., 6., 7.]])
    >>> labels = np.array([[0, 0, 1],
    ...                    [0, 1, 0],
    ...                    [1, 0, 2]])
    >>> naive_grouped_rowwise_apply(data, labels, lambda row: row - row.min())
    array([[ 0.,  1.,  0.],
           [ 0.,  0.,  2.],
           [ 0.,  0.,  0.]])
    >>> naive_grouped_rowwise_apply(data, labels, lambda row: row / row.sum())
    array([[ 0.33333333,  0.66666667,  1.        ],
           [ 0.33333333,  1.        ,  0.66666667],
           [ 1.        ,  1.        ,  1.        ]])
    """
    if out is None:
        out = np.empty_like(data)

    for row, label_row, out_row in zip(data, group_labels, out):
        for label in np.unique(label_row):
            locs = label_row == label
            out_row[locs] = func(row[locs], *func_args)
    return out



import functools
from collections import namedtuple  # noqa: compatibility with python 3.11
from contextlib import contextmanager, ExitStack
from html import escape as escape_html
from math import ceil
from types import MappingProxyType as mappingproxy


def consistent_round(val):
    if (val % 1) >= 0.5:
        return ceil(val)
    else:
        return round(val)


wraps = functools.wraps

unicode = type("")

__all__ = [
    "ExitStack",
    "consistent_round",
    "contextmanager",
    "escape_html",
    "mappingproxy",
    "unicode",
    "wraps",
]



import datetime
from abc import abstractmethod
from dataclasses import dataclass


@dataclass(frozen=True)
class Asset:
    sid: int | None  # if SID is None, then it is a new asset and we want it to have automatically assigned SID
    asset_name: str
    start_date: datetime.date | None
    end_date: datetime.date | None
    first_traded: datetime.date | None
    auto_close_date: datetime.date | None
    mic: str | None

    @abstractmethod
    def get_symbol_by_exchange(self, exchange_name: str | None) -> str | None: ...



from sqlalchemy.orm import Mapped

from ziplime.core.db.annotated_types import StringPK, ExchangeFK
from ziplime.core.db.base_model import BaseModel


class FuturesRootSymbol(BaseModel):
    __tablename__ = "futures_root_symbols"

    root_symbol: Mapped[StringPK]
    root_symbol_id: Mapped[int]
    sector: Mapped[str]
    description: Mapped[str]
    exchange: Mapped[ExchangeFK]
